# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import (
    MultiInputSingleOutputParams,
    NegationParams,
    SingleInputSingleOutputParams,
    StringConstantParams,
)
from kamae.spark.utils import multi_input_single_output_scalar_transform
from kamae.tensorflow.layers import StringContainsLayer

from .base import BaseTransformer


class StringContainsTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    MultiInputSingleOutputParams,
    NegationParams,
    StringConstantParams,
):
    """
    String contains Spark Transformer for use in Spark pipelines.
    This transformer performs a string contains operation on the input column.
    If the string constant is specified, we use it for the string contains
    on the single input. Otherwise, if multiple input columns are specified,
    we check if the first input column contains the second.
    Used for cases where you want to keep the input the same.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        inputCols: Optional[List[str]] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        stringConstant: Optional[str] = None,
        negation: bool = False,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initializes an StringContainsTransformer transformer.

        :param inputCol: Input column name.
        :param inputCols: List of input column names.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column(s) to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param stringConstant: String constant to use in string contains
        operation.
        Only used in single input scenario.
        :param negation: Whether to negate the string contains operation.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(negation=False, stringConstant=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def setInputCols(self, value: List[str]) -> "StringContainsTransformer":
        """
        Overrides setting the input columns for the transformer.
        Throws an error if we do not have exactly two input columns.

        :param value: List of input columns.
        :returns: Class instance.
        """
        if len(value) != 2:
            raise ValueError(
                """When setting inputCols for StringContainsTransformer,
                there must be exactly two input columns."""
            )
        return self._set(inputCols=value)

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which contains the result of the string contains operation.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_cols = self.get_multiple_input_cols(
            constant_param_name="stringConstant", input_cols_limit=2
        )

        # input_cols can contain either actual columns or lit(constants). In order to
        # determine the datatype of the input columns, we select them from the dataset
        # first.
        input_col_names = dataset.select(input_cols).columns
        input_col_datatypes = [
            self.get_column_datatype(dataset=dataset.select(input_cols), column_name=c)
            for c in input_col_names
        ]

        def string_contains(
            x: Column, input_col_names: List[str], negation: bool
        ) -> Column:
            col_expr = F.when(
                x[input_col_names[1]] == F.lit(""), x[input_col_names[0]] == F.lit("")
            ).otherwise(x[input_col_names[0]].contains(x[input_col_names[1]]))
            return col_expr if not negation else ~col_expr

        output_col = multi_input_single_output_scalar_transform(
            input_cols=input_cols,
            input_col_names=input_col_names,
            input_col_datatypes=input_col_datatypes,
            func=lambda x: string_contains(x, input_col_names, self.getNegation()),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the StringContainsLayer transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs a string contains operation.
        """
        return StringContainsLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            negation=self.getNegation(),
            string_constant=self.getStringConstant(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .absolute_value import AbsoluteValueLayer  # noqa: F401
from .array_concatenate import ArrayConcatenateLayer  # noqa: F401
from .array_crop import ArrayCropLayer  # noqa: F401
from .array_split import ArraySplitLayer  # noqa: F401
from .array_subtract_minimum import ArraySubtractMinimumLayer  # noqa: F401
from .bearing_angle import BearingAngleLayer  # noqa: F401
from .bin import BinLayer  # noqa: F401
from .bloom_encode import BloomEncodeLayer  # noqa: F401
from .bucketize import BucketizeLayer  # noqa: F401
from .conditional_standard_scale import ConditionalStandardScaleLayer  # noqa: F401
from .cosine_similarity import CosineSimilarityLayer  # noqa: F401
from .current_date import CurrentDateLayer  # noqa: F401
from .current_date_time import CurrentDateTimeLayer  # noqa: F401
from .current_unix_timestamp import CurrentUnixTimestampLayer  # noqa: F401
from .date_add import DateAddLayer  # noqa: F401
from .date_diff import DateDiffLayer  # noqa: F401
from .date_parse import DateParseLayer  # noqa: F401
from .date_time_to_unix_timestamp import DateTimeToUnixTimestampLayer  # noqa: F401
from .divide import DivideLayer  # noqa: F401
from .exp import ExpLayer  # noqa: F401
from .exponent import ExponentLayer  # noqa: F401
from .hash_index import HashIndexLayer  # noqa: F401
from .haversine_distance import HaversineDistanceLayer  # noqa: F401
from .identity import IdentityLayer  # noqa: F401
from .if_statement import IfStatementLayer  # noqa: F401
from .impute import ImputeLayer  # noqa: F401
from .lambda_function import LambdaFunctionLayer  # noqa: F401
from .list_max import ListMaxLayer  # noqa: F401
from .list_mean import ListMeanLayer  # noqa: F401
from .list_median import ListMedianLayer  # noqa: F401
from .list_min import ListMinLayer  # noqa: F401
from .list_std_dev import ListStdDevLayer  # noqa: F401
from .log import LogLayer  # noqa: F401
from .logical_and import LogicalAndLayer  # noqa: F401
from .logical_not import LogicalNotLayer  # noqa: F401
from .logical_or import LogicalOrLayer  # noqa: F401
from .max import MaxLayer  # noqa: F401
from .mean import MeanLayer  # noqa: F401
from .min import MinLayer  # noqa: F401
from .modulo import ModuloLayer  # noqa: F401
from .multiply import MultiplyLayer  # noqa: F401
from .numerical_if_statement import NumericalIfStatementLayer  # noqa: F401
from .one_hot_encode import OneHotEncodeLayer, OneHotLayer  # noqa: F401
from .ordinal_array_encode import OrdinalArrayEncodeLayer  # noqa: F401
from .round import RoundLayer  # noqa: F401
from .round_to_decimal import RoundToDecimalLayer  # noqa: F401
from .standard_scale import StandardScaleLayer  # noqa: F401
from .string_affix import StringAffixLayer  # noqa: F401
from .string_array_constant import StringArrayConstantLayer  # noqa: F401
from .string_case import StringCaseLayer  # noqa: F401
from .string_concatenate import StringConcatenateLayer  # noqa: F401
from .string_contains import StringContainsLayer  # noqa: F401
from .string_contains_list import StringContainsListLayer  # noqa: F401
from .string_equals_if_statement import StringEqualsIfStatementLayer  # noqa: F401
from .string_index import StringIndexLayer  # noqa: F401
from .string_isin_list import StringIsInListLayer  # noqa: F401
from .string_list_to_string import StringListToStringLayer  # noqa: F401
from .string_map import StringMapLayer  # noqa: F401
from .string_replace import StringReplaceLayer  # noqa: F401
from .string_to_string_list import StringToStringListLayer  # noqa: F401
from .sub_string_delim_at_index import SubStringDelimAtIndexLayer  # noqa: F401
from .subtract import SubtractLayer  # noqa: F401
from .sum import SumLayer  # noqa: F401
from .unix_timestamp_to_date_time import UnixTimestampToDateTimeLayer  # noqa: F401



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import Hashing

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class BloomEncodeLayer(BaseLayer):
    """
    Performs a bloom encoding on the input tensor. Uses multiple hash functions to
    encode the input tensor, significantly reducing the dimensionality of the input
    and also avoiding collisions. See paper for more details.
    https://arxiv.org/pdf/1706.03993.pdf

    In Kamae we actually use the same hash function for all the hash functions,
    but we use a salt to make sure that the hash functions are different. Therefore,
    this can be seen as a psuedo-bloom encoding.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        num_hash_fns: int = 3,
        num_bins: int = None,
        mask_value: Union[int, str] = None,
        feature_cardinality: int = None,
        use_heuristic_num_bins: bool = False,
        **kwargs,
    ) -> None:
        """
        Intialises the BloomEncodeLayer layer.

        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param num_hash_fns: Number of hash functions to use. Defaults to 3.
        The paper suggests a range of 2-4 hash functions for optimal performance.
        :param num_bins: Number of hash bins. Note that this includes the `mask_value`
        bin, so the effective number of bins is `(num_bins - 1)` if `mask_value`
        is set. If `use_heuristic_num_bins` is set to True, then this parameter is
        ignored and the number of bins is automatically set. See the description of this
        parameter below for how the heuristic is built.
        :param mask_value: A value that represents masked inputs, which are mapped to
        index 0. Defaults to None, meaning no mask term will be added and the
        hashing will start at index 0.
        :param feature_cardinality: The cardinality of the input tensor. Needed to use
        the heuristic to set the number of bins. Defaults to None, meaning the number of
        bins will not be set using the heuristic and must be set manually.
        :param use_heuristic_num_bins: If set to True, the number of bins is
        automatically set by fixing the ratio of the feature dimensionality to the
        number of bins to be b/f = 0.2. This ratio was found to be optimal in the paper
        for a wide variety of usecases. Therefore, num_bins = feature_cardinality * 0.2.
        This reduces the cardinality of the input tensor by 5x.
        Requires the `feature_cardinality` parameter to be set. Defaults to False.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if num_hash_fns < 2:
            raise ValueError("The number of hash functions must be at least 2.")
        self.num_hash_fns = num_hash_fns
        self.mask_value = mask_value
        self.feature_cardinality = feature_cardinality
        self.use_heuristic_num_bins = use_heuristic_num_bins

        if use_heuristic_num_bins and feature_cardinality is None:
            raise ValueError(
                """If use_heuristic_num_bins is set to True, then the
                feature_cardinality parameter must be set."""
            )
        if num_bins is None and not use_heuristic_num_bins:
            raise ValueError(
                """If use_heuristic_num_bins is set to False, then the
                num_bins parameter must be set."""
            )
        self.num_bins = (
            num_bins
            if not use_heuristic_num_bins
            else max(round(feature_cardinality * 0.2), 2)
        )
        # We need to create multiple hashing layers if we have a mask_value, as the
        # mask_value needs salting in the same manner as the input tensor. Hence it is
        # not constant across the hash functions. If the mask_value is None, then we
        # can use the same hash function for all the hash functions.
        if mask_value is None:
            hash_fn = Hashing(num_bins=self.num_bins)
            self.hash_fns = {f"{i}": hash_fn for i in range(self.num_hash_fns)}
        else:
            self.hash_fns = {
                f"{i}": Hashing(
                    num_bins=self.num_bins,
                    mask_value=f"{self.mask_value}{i}",
                )
                for i in range(self.num_hash_fns)
            }

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the bloom encoding on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input
        is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to be encoded.
        :returns: Encoded tensor.
        """
        # Expand dimensions to add the bloom encoding dimension for two scenarios:
        # 1. If the final dimension is not 1, in which case we do not want to use
        # this dimension for the encoding.
        # 2. If the rank of the tensor is less than 2, then we have a single dimensional
        # tensor thus we add a dimension for the encoding.
        expanded_inputs = (
            tf.expand_dims(inputs, axis=-1)
            if inputs.shape[-1] != 1 or len(inputs.shape) < 2
            else inputs
        )
        # Salt the inputs to create multiple hash functions
        # Add `i` to the input tensor, where `i` represents the ith hash function.
        salted_inputs = [
            tf.strings.join(
                [expanded_inputs, tf.zeros_like(expanded_inputs)], separator=str(i)
            )
            for i in range(self.num_hash_fns)
        ]
        # Hash the salted inputs.
        hashed_inputs = [
            self.hash_fns[f"{i}"](salted_inputs[i]) for i in range(self.num_hash_fns)
        ]
        return tf.concat(hashed_inputs, axis=-1)

    def get_config(self) -> Dict[str, Any]:
        """
        Returns the configuration of the BloomEncode layer.

        :returns: Configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "num_hash_fns": self.num_hash_fns,
                "num_bins": self.num_bins,
                "mask_value": self.mask_value,
                "feature_cardinality": self.feature_cardinality,
                "use_heuristic_num_bins": self.use_heuristic_num_bins,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides utilities for tensorflow layer inputs"""
from typing import Callable, Iterable, List, Union

import tensorflow as tf

from kamae.tensorflow.typing import Tensor


def iter_values(x) -> Iterable:
    """
    Returns an iterator over the values of a generic iterator.
    Will be used to construct lists from iterables such as lists, tuples, dicts, etc.

    :param x: An iterable
    :returns: An iterator over the values of the iterable.
    """
    if hasattr(x, "itervalues"):
        return x.itervalues()
    if hasattr(x, "values"):
        return iter(x.values())
    return iter(x)


def enforce_single_tensor_input(layer_call_method: Callable) -> Callable:
    """
    Enforces that the inputs to a layer are a single tensor. If the inputs are an
    iterable, then we check it has a single element and that the element is a tensor.
    If the inputs are a tensor, then we return the tensor.

    :param layer_call_method: The layer's call method to decorate.
    :raises TypeError: If the inputs are an iterable with more than one element.
    :returns: The function called with a single tensor.
    """

    def _enforce_single_tensor_input(
        self,
        inputs: Union[Tensor, Iterable[Tensor]],
        **kwargs,
    ) -> Tensor:
        if tf.is_tensor(inputs):
            # If the inputs are a tensor, then we return the tensor.
            processed_inputs = inputs
        else:
            input_list = list(iter_values(inputs))
            if len(input_list) == 1 and tf.is_tensor(input_list[0]):
                # If the inputs are an iterable with a single tensor,
                # then we return the tensor.
                processed_inputs = input_list[0]
            else:
                # Otherwise, we raise an error.
                raise ValueError(
                    f"""Expected inputs to be a single tensor, but got a list of
                    {len(input_list)} tensors."""
                )
        return layer_call_method(self, processed_inputs, **kwargs)

    return _enforce_single_tensor_input


def enforce_multiple_tensor_input(layer_call_method: Callable) -> Callable:
    """
    Enforces that the inputs to a layer are an iterable of tensors.
    We check that all elements are tensors. If the inputs are a single tensor, rather
    than an iterable we raise an error.

    :param layer_call_method: The layer's call method to decorate.
    :raises TypeError: If the inputs are a single tensor, an iterable of length 1
    or an iterable of non-tensors.
    :returns: The function called with a list of tensors.
    """

    def _enforce_multiple_tensor_input(
        self,
        inputs: Union[Tensor, Iterable[Tensor]],
        **kwargs,
    ) -> List[Tensor]:
        if tf.is_tensor(inputs):
            raise ValueError(
                """Expected inputs to be a iterable of tensors,
                but got a single tensor."""
            )
        else:
            input_list = list(iter_values(inputs))
            if len(input_list) > 1 and all(
                [tf.is_tensor(input_tensor) for input_tensor in input_list]
            ):
                processed_inputs = input_list
            else:
                raise ValueError(
                    """Invalid inputs. Expected inputs to be an iterable of tensors,
                    but either got an iterable of non-tensors or a single tensor."""
                )
        return layer_call_method(self, processed_inputs, **kwargs)

    return _enforce_multiple_tensor_input


def allow_single_or_multiple_tensor_input(layer_call_method: Callable) -> Callable:
    """
    Enforces that the inputs to a layer are either a single tensor or a list of tensors.
    If the inputs are an iterable, then we check that all elements are tensors. If the
    inputs are a tensor, then we return a list containing the tensor.

    :param layer_call_method: The layer's call method to decorate.
    :returns: The function called with a list of tensors.
    """

    def _allow_single_or_multiple_tensor_input(
        self,
        inputs: Union[Tensor, Iterable[Tensor]],
        **kwargs,
    ) -> List[Tensor]:
        if tf.is_tensor(inputs):
            processed_inputs = [inputs]
        else:
            input_list = list(iter_values(inputs))
            if all([tf.is_tensor(input_tensor) for input_tensor in input_list]):
                processed_inputs = input_list
            else:
                raise ValueError(
                    """All elements of the inputs must be tensors, but got an iterable
                    containing non-tensors."""
                )
        return layer_call_method(self, processed_inputs, **kwargs)

    return _allow_single_or_multiple_tensor_input



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import SingleInputSingleOutputParams, StringIndexParams
from kamae.spark.transformers import StringIndexTransformer
from kamae.spark.utils import collect_labels_array

from .base import BaseEstimator


class StringIndexEstimator(
    BaseEstimator,
    SingleInputSingleOutputParams,
    StringIndexParams,
):
    """
    String indexer Spark Estimator for use in Spark pipelines.
    This estimator is used to collect all the string labels for a given column.
    When fit is called it returns a StringIndexerLayerModel which can be used
    to index additional feature columns using the same string labels.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        stringOrderType: Optional[str] = None,
        maskToken: Optional[str] = None,
        numOOVIndices: int = 1,
        maxNumLabels: Optional[int] = None,
    ) -> None:
        """
        Initializes the StringIndexEstimator estimator.
        Sets all parameters to given inputs.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param: stringOrderType: How to order the string indices.
        Options are 'frequencyAsc', 'frequencyDesc', 'alphabeticalAsc',
        'alphabeticalDesc'.
        :param maskToken: Token to use for masking.
        If set, the token will be indexed as 0.
        :param numOOVIndices: Number of out of vocabulary indices to use.
        :param maxNumLabels: Optional value to limit the size of the vocabulary.
        Default is 1.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            stringOrderType="frequencyDesc",
            numOOVIndices=1,
            maskToken=None,
            maxNumLabels=None,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _fit(self, dataset: DataFrame) -> "StringIndexTransformer":
        """
        Fits the StringIndexEstimator estimator to the given dataset.
        Returns a StringIndexerLayerModel which can be used to index columns using
        the collected string labels.

        :param dataset: Pyspark dataframe to fit the estimator to.
        :returns: StringIndexerLayerModel instance with collected string labels.
        """
        column_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        labels = collect_labels_array(
            dataset=dataset,
            column=F.col(self.getInputCol()),
            column_datatype=column_datatype,
            string_order_type=self.getStringOrderType(),
            mask_token=self.getMaskToken(),
            max_num_labels=self.getMaxNumLabels(),
        )
        self.setLabelsArray(labels)

        return StringIndexTransformer(
            inputCol=self.getInputCol(),
            outputCol=self.getOutputCol(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            labelsArray=self.getLabelsArray(),
            stringOrderType=self.getStringOrderType(),
            numOOVIndices=self.getNumOOVIndices(),
            maskToken=self.getMaskToken(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType

from kamae.spark.params import SingleInputMultiOutputParams
from kamae.spark.utils import single_input_single_output_array_transform
from kamae.tensorflow.layers import ArraySplitLayer

from .base import BaseTransformer


class ArraySplitTransformer(
    BaseTransformer,
    SingleInputMultiOutputParams,
):
    """
    ArraySplit Spark Transformer for use in Spark pipelines.
    This transformer splits an array column into multiple columns.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCols: Optional[List[str]] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initialize a ArraySplitTransformer transformer.

        :param inputCol: Input column name.
        :param outputCols: Output column names.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column(s) to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column for each output column equal
        to the value of the input column at the given index.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        output_cols = []
        for i, column in enumerate(self.getOutputCols()):
            output_col = single_input_single_output_array_transform(
                input_col=F.col(self.getInputCol()),
                input_col_datatype=input_datatype,
                func=lambda x: F.element_at(x, i + 1),
            )
            output_cols.append(output_col.alias(self.getOutputCols()[i]))
        original_columns = [F.col(c) for c in dataset.columns]
        select_cols = original_columns + output_cols
        return dataset.select(select_cols)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for that unstacks the input tensor and reshapes
        to the original shape.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that slices the input tensors.
        """
        return ArraySplitLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            axis=-1,
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np
import pandas as pd
import tensorflow as tf

from kamae.sklearn.params import SingleInputSingleOutputMixin
from kamae.tensorflow.layers import LogLayer

from .base import BaseTransformer


class LogTransformer(BaseTransformer, SingleInputSingleOutputMixin):
    """
    Log Scikit-Learn Transformer for use in Scikit-Learn pipelines.
    This transformer applies a log(alpha + x) transform to the input column.
    """

    def __init__(
        self,
        input_col: str,
        output_col: str,
        layer_name: str,
        alpha: Optional[float] = None,
    ) -> None:
        """
        Intializes a LogTransformLayer transformer. Sets the default values of:

        - alpha: 1

        :param input_col: Input column name.
        :param output_col: Output column name.
        :param layer_name: Name of the layer. Used as the name of the tensorflow layer
        :param alpha: Value to use in log transform: log(alpha + x). Default is 1.
        :returns: None - class instantialized.
        """
        super().__init__()
        self.input_col = input_col
        self.output_col = output_col
        self.layer_name = layer_name
        self.alpha = float(alpha) if alpha is not None else 1.0

    def fit(self, X: pd.DataFrame, y=None) -> "LogTransformer":
        """
        Fits the transformer. Does nothing since this is just a transformer.

        :param X: Pandas dataframe to fit the transformer to.
        :param y: Not used, present here for API consistency by convention.
        :returns: Fit pipeline, in this case the transformer itself.
        """
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transforms the data using the transformer. Creates a new column with name
        `output_col`, which applies log(alpha + x) transform to the `input_col`.

        :param X: Pandas dataframe to transform.
        :param y: Not used, present here for API consistency by convention.
        :returns: Transformed data.
        """
        X[self.output_col] = np.log(X[self.input_col] + self.alpha)
        return X

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the log transform.

        :returns: Tensorflow keras layer with name equal to the layerName parameter
        that performs the log(alpha + x) operation.
        """
        alpha = self.alpha
        return LogLayer(name=self.layer_name, alpha=alpha)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .absolute_value import AbsoluteValueTransformer  # noqa: F401
from .array_concatenate import ArrayConcatenateTransformer  # noqa: F401
from .array_crop import ArrayCropTransformer  # noqa: F401
from .array_split import ArraySplitTransformer  # noqa: F401
from .array_subtract_minimum import ArraySubtractMinimumTransformer  # noqa: F401
from .base import BaseTransformer  # noqa: F401
from .bearing_angle import BearingAngleTransformer  # noqa: F401
from .bin import BinTransformer  # noqa: F401
from .bloom_encode import BloomEncodeTransformer  # noqa: F401
from .bucketize import BucketizeTransformer  # noqa: F401
from .conditional_standard_scale import (  # noqa: F401
    ConditionalStandardScaleTransformer,
)
from .cosine_similarity import CosineSimilarityTransformer  # noqa: F401
from .current_date import CurrentDateTransformer  # noqa: F401
from .current_date_time import CurrentDateTimeTransformer  # noqa: F401
from .current_unix_timestamp import CurrentUnixTimestampTransformer  # noqa: F401
from .date_add import DateAddTransformer  # noqa: F401
from .date_diff import DateDiffTransformer  # noqa: F401
from .date_parse import DateParseTransformer  # noqa: F401
from .date_time_to_unix_timestamp import (  # noqa: F401
    DateTimeToUnixTimestampTransformer,
)
from .divide import DivideTransformer  # noqa: F401
from .exp import ExpTransformer  # noqa: F401
from .exponent import ExponentTransformer  # noqa: F401
from .hash_index import HashIndexTransformer  # noqa: F401
from .haversine_distance import HaversineDistanceTransformer  # noqa: F401
from .identity import IdentityTransformer  # noqa: F401
from .if_statement import IfStatementTransformer  # noqa: F401
from .impute import ImputeTransformer  # noqa: F401
from .lambda_function import LambdaFunctionTransformer  # noqa: F401
from .list_max import ListMaxTransformer  # noqa: F401
from .list_mean import ListMeanTransformer  # noqa: F401
from .list_median import ListMedianTransformer  # noqa: F401
from .list_min import ListMinTransformer  # noqa: F401
from .list_std_dev import ListStdDevTransformer  # noqa: F401
from .log import LogTransformer  # noqa: F401
from .logical_and import LogicalAndTransformer  # noqa: F401
from .logical_not import LogicalNotTransformer  # noqa: F401
from .logical_or import LogicalOrTransformer  # noqa: F401
from .max import MaxTransformer  # noqa: F401
from .mean import MeanTransformer  # noqa: F401
from .min import MinTransformer  # noqa: F401
from .modulo import ModuloTransformer  # noqa: F401
from .multiply import MultiplyTransformer  # noqa: F401
from .numerical_if_statement import NumericalIfStatementTransformer  # noqa: F401
from .one_hot_encode import OneHotEncodeTransformer  # noqa: F401
from .ordinal_array_encode import OrdinalArrayEncodeTransformer  # noqa: F401
from .round import RoundTransformer  # noqa: F401
from .round_to_decimal import RoundToDecimalTransformer  # noqa: F401
from .shared_one_hot_encode import SharedOneHotEncodeTransformer  # noqa: F401
from .shared_string_index import SharedStringIndexTransformer  # noqa: F401
from .standard_scale import StandardScaleTransformer  # noqa: F401
from .string_affix import StringAffixTransformer  # noqa: F401
from .string_array_constant import StringArrayConstantTransformer  # noqa: F401
from .string_case import StringCaseTransformer  # noqa: F401
from .string_concatenate import StringConcatenateTransformer  # noqa: F401
from .string_contains import StringContainsTransformer  # noqa: F401
from .string_contains_list import StringContainsListTransformer  # noqa: F401
from .string_equals_if_statement import StringEqualsIfStatementTransformer  # noqa: F401
from .string_index import StringIndexTransformer  # noqa: F401
from .string_isin_list import StringIsInListTransformer  # noqa: F401
from .string_list_to_string import StringListToStringTransformer  # noqa: F401
from .string_map import StringMapTransformer  # noqa: F401
from .string_replace import StringReplaceTransformer  # noqa: F401
from .string_to_string_list import StringToStringListTransformer  # noqa: F401
from .sub_string_delim_at_index import SubStringDelimAtIndexTransformer  # noqa: F401
from .subtract import SubtractTransformer  # noqa: F401
from .sum import SumTransformer  # noqa: F401
from .unix_timestamp_to_date_time import (  # noqa: F401
    UnixTimestampToDateTimeTransformer,
)



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Tuple, Union

import keras
import networkx as nx
import tensorflow as tf
from packaging.version import Version

from kamae.tensorflow.layers import IdentityLayer

keras_version = Version(keras.__version__)


class PipelineGraph:
    """
    PipelineGraph is a class that constructs a graph of the pipeline stages.
    This is used to determine the order in which the layers should be constructed.
    The graph is built by adding edges between layers that have the same input column
    as the output column of a previous layer. If the input is not an output of any other
    layer, then it is assumed to be an input to the pipeline.

    The graph is then topologically sorted to determine the order in which the layers
    should be constructed. Iterating through this order, the layers are constructed by
    calling the get_tf_layer method of each stage. The inputs to the layer are
    determined by the outputs of the previous layers.
    """

    def __init__(self, stage_dict: Dict[str, Any]) -> None:
        """
        Initialize the PipelineGraph class with a dictionary of stage information.

        :param stage_dict: Dictionary of stages to add to the graph.
        :returns: None - class instance is initialized.
        """
        self.stage_dict = stage_dict
        self.graph = self.add_stage_edges(nx.DiGraph())
        self.transform_order = [
            node for node in nx.topological_sort(self.graph) if node in self.stage_dict
        ]

        # We keep a dictionary of layers to keep track of which have been reused.
        # This allows us to easily get the output layers
        self.layer_store = {}
        self.inputs = {}

    def update_layer_store_with_key(self, layer_key, layer_output) -> None:
        """
        Updates the layer store at a specific key with the layer output and whether
        it was reused. A layer is deemed to be reused if it is already present in
        the layer store.

        :param layer_key: Key to update the layer store with.
        :param layer_output: Layer output to update.
        :returns: None - layer store is updated.
        """
        if layer_key in self.layer_store:
            self.layer_store[layer_key] = {"output": layer_output, "reused": True}
        else:
            self.layer_store[layer_key] = {"output": layer_output, "reused": False}

    def update_layer_store(self, layer_dict: Dict[str, tf.Tensor]) -> None:
        """
        Given a dictionary of layer output names and tensor outputs,
        update the layer store.

        :param layer_dict: Dictionary of layer names and outputs.
        :returns: None - layer store is updated.
        """
        for name, output in layer_dict.items():
            self.update_layer_store_with_key(layer_key=name, layer_output=output)

    def get_layer_output_from_layer_store(self, layer_output_name: str) -> tf.Tensor:
        """
        Given a layer name and index, get the output from the layer store.

        :param layer_output_name: Name of the layer output
        :returns: Tensor output of the layer
        """
        return self.layer_store[layer_output_name]["output"]

    def add_stage_edges(self, graph):
        """
        Add edges to the graph based on the inputs and outputs of each stage.
        Specifically for each stage, we connect the inputs of a stage to itself and the
        stage to its outputs.

        :param graph: NetworkX DAG to add edges to.
        :returns: Graph with edges added.
        """
        edges_to_add = []
        for layer_name, layer_info in self.stage_dict.items():
            # Add edges for all inputs
            edges_to_add.extend(
                [(input_name, layer_name) for input_name in layer_info["inputs"]]
            )
            # Add edges for all outputs
            edges_to_add.extend(
                [(layer_name, output_name) for output_name in layer_info["outputs"]]
            )

        graph.add_edges_from(edges_to_add)
        return graph

    def get_model_outputs(
        self, output_names: Optional[List[str]] = None
    ) -> Dict[str, tf.Tensor]:
        """
        Gets the outputs of the model. If output_names is provided, we use this to find
        the outputs for the model. Otherwise, the outputs are those that are not reused
        and not inputs. We also apply an identity layer to the outputs, so we
        can rename them with the same name as the output columns of the layer.

        :param output_names: Optional list of output names. If provided, the outputs
        are only allowed to be within this list.
        :returns: Dictionary of model output tensors.
        """
        if output_names is None:
            # If no specified output names then these are outputs that are not reused
            # and not inputs.
            output_names = [
                k
                for k, v in self.layer_store.items()
                if not v["reused"] and k not in self.inputs
            ]
        return {
            # Do not wrap with identity if we are just passing through an input.
            k: IdentityLayer(name=k)(v["output"])
            if k not in self.inputs
            else v["output"]
            for k, v in self.layer_store.items()
            if k in output_names
        }

    def build_keras_inputs(
        self, tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]]
    ) -> None:
        """
        Builds a Keras input layer for the given node.

        Specifically, we get a single node from the out edges of the current node.
        The out edges are a tuple of edges `(u, v)` where `u` is the current node and
        `v` is a node takes the current node (`u`) as input.

        Using `v` we can get the input dimension and type and therefore construct the
        keras input layer. We then store this layer as an input and update the
        layer store.

        :param tf_input_schema: List of tf.TypeSpec objects containing the input schema
        for the model or a list of dict config to be passed to the Input constructor.
        :returns: None - layer store is updated and input layer is stored in the
        inputs dict.
        """

        if isinstance(tf_input_schema, list) and all(
            isinstance(x, tf.TypeSpec) for x in tf_input_schema
        ):
            if keras_version >= Version("3.0.0"):
                raise ValueError(
                    "tf.TypeSpec is not supported in Keras 3, please use a dict config"
                )
            input_config = [
                {
                    "name": spec.name,
                    "type_spec": spec,
                }
                for spec in tf_input_schema
            ]
        elif isinstance(tf_input_schema, list) and all(
            isinstance(x, dict) for x in tf_input_schema
        ):
            input_config = tf_input_schema
        else:
            raise ValueError("tf_input_schema must be a list of tf.TypeSpec or dict!")

        for conf in input_config:
            name = conf.get("name", None)
            if name is None:
                raise ValueError(
                    "Input schema must have names for all inputs, but found None"
                )
            input_layer = tf.keras.layers.Input(**conf)
            self.inputs[name] = input_layer
            self.update_layer_store_with_key(layer_key=name, layer_output=input_layer)

    def sort_inputs(
        self, layer_name: str, input_dict: Dict[str, tf.Tensor]
    ) -> List[tf.Tensor]:
        """
        Sorts the inputs for a given layer based on the order of the inputs in the
        stage dict. This is needed because layers with multiple inputs are not
        guaranteed to be in the correct order when built from the graph as the
        topological sort can be different to the order in the stage dict.

        :param layer_name: Name of the layer
        :param input_dict: Dictionary of inputs for the layer.
        :returns: Sorted list of inputs for the layer.
        """
        stage_inputs = self.stage_dict[layer_name]["inputs"]
        return [input_dict[i] for i in stage_inputs]

    def build_transform_layer_inputs(
        self, node: str, in_edges: List[Tuple[str, str]]
    ) -> List[tf.Tensor]:
        """
        Constructs all the layers that are connected to the current node.
        These are either input layers or the outputs of previous layers.

        The in edges are a tuple of edges `(u, v)` where `v` is the current node and
        `u` is a node that is an input to the current node (`v`).

        Using `u` we can tell if this is the output of another transformation or an
        input layer. If it's an input layer, we retrieve it from the inputs dictionary.
        If it's the output of another transformation, we retrieve it from the
        layer store.

        :param node: Current node (name of layer).
        :param in_edges: List of in edges for the current node.
        :returns: List of layer inputs for the current node/layer.
        """
        # Here we get all layer outputs that are connected to this node.
        # We need these so we can apply the current node's layer to
        # the output of the previous layers.
        # Since we topologically sorted the nodes,
        # all previous layers will have already been created.

        # Get the in edge node names
        in_edge_node_names = [in_edge[0] for in_edge in in_edges]
        # For each in edge, find the output that maps to this node.
        layer_output_from_in_edge = [
            (stage_name, in_edge_node_name)
            for in_edge_node_name in in_edge_node_names
            for stage_name, stage_info in self.stage_dict.items()
            if in_edge_node_name in stage_info["outputs"]
        ]
        # Get any input layers that are connected to the node via the in edges
        input_layers_from_in_edge = {
            name: layer
            for name, layer in self.inputs.items()
            if name in in_edge_node_names
        }

        # For each layer output in edge,
        # get its corresponding layer output from the layer store.
        # Update the store to indicate we have reused this layer.
        # Thus, it is not an output layer
        in_edge_layers_inputs = {}
        for name, output_name in layer_output_from_in_edge:
            layer = self.get_layer_output_from_layer_store(output_name)
            self.update_layer_store_with_key(layer_key=output_name, layer_output=layer)
            in_edge_layers_inputs[output_name] = layer

        # Sort the inputs according to the order set in the Spark transformers
        input_dict = {**in_edge_layers_inputs, **input_layers_from_in_edge}
        layer_inputs = self.sort_inputs(
            layer_name=node,
            input_dict=input_dict,
        )
        return layer_inputs

    @staticmethod
    def override_hyperparameters(
        layer: Union[tf.keras.layers.Layer, List[tf.keras.layers.Layer]],
        hp_override: Dict[str, Any] = None,
    ) -> Union[tf.keras.layers.Layer, List[tf.keras.layers.Layer]]:
        """
        Overrides layer arguments with hyperparameters provided in the
        hyperparameter override dictionary.

        :param layer: Layer to override hyperparameters for.
        :param hp_override: Optional dictionary of hyperparameters to override.
        :returns: Layer with hyperparameters overridden.
        """

        def update_layer(
            layer: tf.keras.layers.Layer, hp_override: Dict[str, Any]
        ) -> tf.keras.layers.Layer:
            config = layer.get_config()
            config.update(hp_override)
            updated_layer = type(layer).from_config(config)
            return updated_layer

        if hp_override is None:
            return layer
        elif isinstance(layer, list):
            overriden_layer = []
            for layer_elem in layer:
                overriden_layer.append(update_layer(layer_elem, hp_override))
            return overriden_layer
        else:
            overriden_layer = update_layer(layer, hp_override)
            return overriden_layer

    def build_keras_transform_layer(
        self,
        node: str,
        in_edges: List[Tuple[str, str]],
        hp_override: Dict[str, Any] = None,
    ) -> None:
        """
        Builds a Keras transformation layer for the given node.
        Gets the layer inputs using the in edges and then applies the layer to
        the inputs. Updates the layer store.

        :param node: Current node (name of layer).
        :param in_edges: List of in edges for the current node.
        :param hp_override: Optional dictionary of hyperparameters to override.
        Used for building Keras tuner model builder functions.
        :returns: None - layer store is updated.
        """
        layer_inputs = self.build_transform_layer_inputs(node=node, in_edges=in_edges)
        layer = self.stage_dict[node]["layer"]
        layer = self.override_hyperparameters(layer=layer, hp_override=hp_override)

        if isinstance(layer, list):
            # If we have a list of layers, we assume that each layer needs to operate
            # on the corresponding input idx in the list of inputs.
            layer_outputs = [
                layer_elem(layer_input)
                for layer_elem, layer_input in zip(layer, layer_inputs)
            ]

        else:
            # If we have a single layer, we assume that it needs to operate on all
            # the inputs.
            layer_outputs = (
                layer(layer_inputs) if len(layer_inputs) > 1 else layer(*layer_inputs)
            )

        layer_output_names = self.stage_dict[node]["outputs"]
        # Make layer outputs a list if it isn't already
        layer_outputs = (
            layer_outputs if isinstance(layer_outputs, list) else [layer_outputs]
        )

        # Zip the output names with the ouputs themselves
        layer_outputs_with_name = {
            layer_output_name: layer_output
            for layer_output_name, layer_output in zip(
                layer_output_names, layer_outputs
            )
        }

        self.update_layer_store(layer_dict=layer_outputs_with_name)

    @staticmethod
    def get_keras_hyperparam_from_config(hp, config):
        """
        Returns a keras hyperparameter object from a config dictionary.

        :param hp: keras_tuner.HyperParameters object passed through from the
        keras tuner model builder function.
        :param config: Config supplied by the user.
        :returns: keras_tuner.HyperParameters method for the given config.
        """

        method_dict = {
            "int": hp.Int,
            "choice": hp.Choice,
            "boolean": hp.Boolean,
            "float": hp.Float,
            "fixed": hp.Fixed,
        }

        hyperparam_dict = {}
        for hyperparams in config:
            try:
                method = method_dict[hyperparams["method"].lower()]
            except KeyError:
                raise ValueError(
                    f"""
                    Hyperparameter method {hyperparams['method']} not supported.
                    Must be one of {",".join(list(method_dict.keys()))}"""
                )

            hyperparam_dict.update(
                {hyperparams["arg_name"]: method(**hyperparams["kwargs"])}
            )
        return hyperparam_dict

    def get_keras_tuner_model_builder(
        self,
        tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]],
        hp_dict: Dict[str, List[Dict[str, Any]]],
        output_names: Optional[List[str]] = None,
    ):
        """
        Returns a Keras tuner model builder function for the current graph.
        This allows the user to tune the hyperparameters of the preprocessing model.
        Useful for scenarios where the best preprocessing variables are not known
        a priori. For example, the num_bins to use for a HashIndexLayer.

        :param tf_input_schema: List of tf.TypeSpec objects containing the input schema
        for the model. Specifically the name, shape and dtype of each input.
        These will be passed as is to the Keras Input layer.
        :param hp_dict: Dictionary of possible hyperparameters for each layer.
        Should be of the format:
        {
            "<LAYER_NAME>": [
                {
                    "arg_name": <NAME_OF_LAYER_ARGUMENT>,
                    "method": <NAME_OF_KERAS_HYPERPARAMETER_METHOD>, e.g. "choice"
                    "kwargs": {
                        <KWARGS_TO_PASS_TO_KERAS_HYPERPARAMETER_METHOD>
                    }
                }
            ]
        }
        :param output_names: Optional list of output names for the Keras model. If
        provided, only the outputs specified are used as model outputs.
        :returns: Model builder function that takes a keras_tuner.HyperParameters class
        and returns a model.
        """

        transform_order = self.transform_order

        def keras_model_builder(hp):
            # We need to clear the layer store and inputs each time we build a model.
            self.layer_store = {}
            self.inputs = {}
            # Build the input layers
            self.build_keras_inputs(tf_input_schema=tf_input_schema)

            for node in transform_order:
                in_edges = list(self.graph.in_edges(node))

                # Try and get the hyperparameter override.
                try:
                    hp_override = self.get_keras_hyperparam_from_config(
                        hp, hp_dict[node]
                    )
                except KeyError:
                    hp_override = None
                self.build_keras_transform_layer(
                    node=node, in_edges=in_edges, hp_override=hp_override
                )

            sorted_inputs = [self.inputs[k] for k in sorted(self.inputs)]
            return tf.keras.Model(
                inputs=sorted_inputs,
                outputs=self.get_model_outputs(output_names=output_names),
            )

        return keras_model_builder

    def build_keras_model(
        self,
        tf_input_schema: Union[List[tf.TypeSpec], List[Dict[str, Any]]],
        output_names: Optional[List[str]] = None,
    ) -> tf.keras.Model:
        """
        Builds a Keras model from the graph.

        :param tf_input_schema: List of tf.TypeSpec objects containing the input schema
        for the model. Each TypeSpec object must define a unique `name` attribute.
        These will be passed as is to the Keras Input layer.
        :param output_names: Optional list of output names for the Keras model. If
        provided, only the outputs specified are used as model outputs.
        :returns: Keras model to be applied to a tensors dictionary: {name: Tensor}.
        """
        # Build the input layers
        self.build_keras_inputs(tf_input_schema=tf_input_schema)

        for node in self.transform_order:
            in_edges = list(self.graph.in_edges(node))
            self.build_keras_transform_layer(node=node, in_edges=in_edges)

        # All the layers are now stored in the layer store,
        # with all inputs/outputs specified.
        # We can now build the model by specifying the inputs and outputs.
        sorted_inputs = {k: self.inputs[k] for k in sorted(self.inputs)}
        return tf.keras.Model(
            inputs=sorted_inputs,
            outputs=self.get_model_outputs(output_names=output_names),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import (
    datetime_day,
    datetime_day_of_year,
    datetime_hour,
    datetime_millisecond,
    datetime_minute,
    datetime_month,
    datetime_second,
    datetime_weekday,
    datetime_year,
    enforce_single_tensor_input,
)

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class DateParseLayer(BaseLayer):
    """
    Parses a date(time) string from yyyy-MM-dd (HH:mm:ss.SSS) format
    into a specified date part tensor.

    Date parts can be one of the following:
    - `DayOfWeek` - day of week (Monday = 1, Sunday = 7)
    - `DayOfMonth` - day of month
    - `DayOfYear` - day of year e.g. (2021-01-01 = 1, 2021-12-31 = 365)
    - `MonthOfYear` - month of year
    - `Year` - year
    - `Hour` - hour e.g. (2021-01-01 00:00:00 = 0, 2021-01-01 23:59:59 = 23)
    - `Minute` - minute e.g. (2021-01-01 00:00:00 = 0, 2021-01-01 00:59:00 = 59)
    - `Second` - second e.g. (2021-01-01 00:00:00 = 0, 2021-01-01 00:00:59 = 59)
    - `Millisecond` - millisecond (2021-01-01 00:00:00.357 = 357)

    In the case a timestamp is not provided, all hour, minutes, seconds and milliseconds
    fields will be returned as 0.

    All date parts except seconds and milliseconds are returned as int32, but due to the
    precision of seconds and milliseconds, these are returned as int64 to prevent
    overflow.

    WARNING: Dates are not checked for validity, so if you pass in a date such
    as "2020-02-30" no errors will be thrown and you will get a nonsense output.
    """

    def __init__(
        self,
        date_part: str,
        name=None,
        input_dtype: str = None,
        output_dtype: str = None,
        default_value: Optional[int] = None,
        **kwargs,
    ) -> None:
        """
        Initialises an instance of the DateParseLayer layer.

        :param date_part: Date part to extract from date.
        :param name: Name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param default_value: Default value to use when the date is the empty string.
        Empty strings can be used when the date is not available.
        :returns: None - class instantiated.
        """
        self.allowed_date_parts = {
            "DayOfWeek",
            "DayOfMonth",
            "DayOfYear",
            "MonthOfYear",
            "Year",
            "Hour",
            "Minute",
            "Second",
            "Millisecond",
        }
        if date_part not in self.allowed_date_parts:
            raise ValueError(f"date_part must be one of {self.allowed_date_parts}")
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.date_part = date_part
        self.default_value = default_value

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.string]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Extracts date part from date(time) string.

        Decorated with `@enforce_single_tensor_input` to ensure that only a single
        tensor is passed in. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Tensor of date(time) strings in the yyyy-MM-dd (HH:mm:ss.SSS)
        format.
        :returns: Date part tensor.

        WARNING: Dates are not checked for validity, so if you pass in a date such
        as "2020-02-30" no errors will be thrown and you will get a nonsense output.
        """
        if self.default_value is not None:
            # Trick to replace empty strings with a valid dummy date, that we ignore
            # later. Otherwise, the parse_date function will raise an error
            replaced_date = tf.where(
                tf.equal(inputs, ""), "2000-01-01 00:00:00.000", inputs
            )
            outputs = tf.where(
                tf.equal(inputs, ""),
                tf.constant(self.default_value, dtype=tf.int64),
                self._parse_date(replaced_date, self.date_part),
            )
        else:
            outputs = self._parse_date(inputs, self.date_part)
        return outputs

    @staticmethod
    def _parse_date(date_tensor: Tensor, date_part: str) -> Tensor:
        """
        Parse date(time) string into a dictionary of date part tensors.

        :param date_tensor: Tensor of date(time) strings in the
        YYYY-mm-dd (HH:MM:ss.SSS) format.
        :returns: Dictionary of date part tensors.
        """

        date_part_functions = {
            "DayOfWeek": datetime_weekday,
            "DayOfMonth": datetime_day,
            "DayOfYear": datetime_day_of_year,
            "MonthOfYear": datetime_month,
            "Year": datetime_year,
            "Hour": datetime_hour,
            "Minute": datetime_minute,
            "Second": datetime_second,
            "Millisecond": datetime_millisecond,
        }

        try:
            return date_part_functions[date_part](date_tensor)
        except KeyError:
            raise ValueError(
                f"""date_part must be one of {list(date_part_functions.keys())}"""
            )

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the DateParse layer.
        Used for saving and loading from a model.

        Specifically adds the `date_part` to the config dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {"date_part": self.date_part, "default_value": self.default_value}
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List

import pyspark.sql.functions as F
from pyspark.sql import Column
from pyspark.sql.types import ArrayType, DataType

from kamae.spark.utils import (
    broadcast_scalar_column_to_array,
    build_udf_return_type,
    get_array_nesting_level,
    nested_arrays_zip,
    nested_lambda,
    nested_transform,
)


def _single_input_single_output_transform(
    input_col: Column,
    input_col_datatype: DataType,
    func: Callable[[Column], Column],
    scalar: bool,
) -> Column:
    """
    Applies either a scalar or array Spark function to a single input column and returns
    a single output column. Ideally this function should not be used directly, the user
    should choose one of the below scalar vs array functions for their usecase. As every
    transform is either scalar or array, not both. This function is used to reduce code
    duplication between the scalar and array functions.

    :param input_col: Input column.
    :param input_col_datatype: Input column datatype.
    :param func: Function to apply to the input column.
    :param scalar: Whether the function to be applied is a scalar function or an
    array function.
    :returns: Output column.
    """
    if not scalar and not isinstance(input_col_datatype, ArrayType):
        raise ValueError(
            f"""Input column was expected to be an array but received datatype
            {input_col_datatype}
            """
        )
    nested_level = get_array_nesting_level(input_col_datatype)
    if not scalar:
        # If the function is not scalar then it operates on the full array, therefore
        # we reduce the nested level by 1.
        nested_level -= 1
    nested_transform_func = nested_transform(func, nested_level)
    return nested_transform_func(input_col)


def single_input_single_output_scalar_transform(
    input_col: Column, input_col_datatype: DataType, func: Callable[[Column], Column]
) -> Column:
    """
    Applies a scalar Spark function (e.g. a Spark standard library function that can be
    applied elementwise to arrays) to a single input column and returns a single output
    column. Caters for the case, where the input column is:

    1. A scalar.
    2. A (possibly nested) array.

    If the input is a scalar, apply `func` directly to the input column.
    If the input is an array, apply `func` elementwise to the input column.

    :param input_col: Input column.
    :param input_col_datatype: Input column datatype.
    :param func: Function to apply to the input column.
    :returns: Output column.
    """
    return _single_input_single_output_transform(
        input_col, input_col_datatype, func, scalar=True
    )


def single_input_single_output_array_transform(
    input_col: Column, input_col_datatype: DataType, func: Callable[[Column], Column]
) -> Column:
    """
    Applies an array Spark function (e.g. a Spark standard library function that
    operates on an array directly, as opposed to elementwise on arrays) to a single
    input column and returns a single output column. Caters for the case, where the
    input column is a nested array, in which case the function is applied to the
    innermost array.

    :param input_col: Input column.
    :param input_col_datatype: Input column datatype.
    :param func: Function to apply to the input column.
    :returns: Output column.
    """
    return _single_input_single_output_transform(
        input_col, input_col_datatype, func, scalar=False
    )


def _single_input_single_output_udf_transform(
    input_col: Column,
    input_col_datatype: DataType,
    func: Callable,
    udf_return_element_datatype: DataType,
    scalar: bool,
) -> Column:
    """
    Applies a Python function (e.g. a lambda function that we will wrap into a
    UDF) to a single input column and returns a single output column. Ideally this
    function should not be used directly, the user should choose one of the below scalar
    vs array functions for their usecase. As every transform is either scalar or array,
    not both. This function is used to reduce code duplication between the scalar and
    array functions.

    :param input_col: Input column.
    :param input_col_datatype: Input column datatype.
    :param func: Function to apply to the input column.
    :param udf_return_element_datatype: Datatype of the UDF return type. Should be the
    raw underlying type.
    :param scalar: Whether the function to be applied is a scalar function or an
    array function.
    :returns: Output column.
    """
    if not scalar and not isinstance(input_col_datatype, ArrayType):
        raise ValueError(
            f"""Input column was expected to be an array but received datatype
            {input_col_datatype}
            """
        )
    nested_level = get_array_nesting_level(input_col_datatype)
    udf_return_type = build_udf_return_type(
        element_dtype=udf_return_element_datatype, nest_level=nested_level
    )
    if not scalar:
        # If the function is not scalar then it operates on the full array, therefore
        # we reduce the nested level by 1.
        nested_level -= 1

    nested_lambda_func = nested_lambda(
        func=func,
        nest_level=nested_level,
    )
    udf_func = F.udf(nested_lambda_func, udf_return_type)
    return udf_func(input_col)


def single_input_single_output_scalar_udf_transform(
    input_col: Column,
    input_col_datatype: DataType,
    func: Callable,
    udf_return_element_datatype: DataType,
) -> Column:
    """
    Applies a scalar Python function (e.g. a lambda function that we will wrap into a
    UDF) to a single input column and returns a single output column. Caters for the
    case, where the input column is:

    1. A scalar.
    2. A (possibly nested) array.

    If the input is a scalar, apply `func` directly to the input column.
    If the input is an array, apply `func` elementwise to the input column.

    `func` should be provided as a python function, and not wrapped in the Spark UDF
    wrapper.

    :param input_col: Input column.
    :param input_col_datatype: Input column datatype.
    :param func: Function to apply to the input column.
    :param udf_return_element_datatype: Datatype of the UDF return type. Should be the
    raw underlying type.
    :returns: Output column.
    """
    return _single_input_single_output_udf_transform(
        input_col, input_col_datatype, func, udf_return_element_datatype, scalar=True
    )


def single_input_single_output_array_udf_transform(
    input_col: Column,
    input_col_datatype: DataType,
    func: Callable,
    udf_return_element_datatype: DataType,
) -> Column:
    """
    Applies an array Python function (e.g. a lambda function that operates on an array
    directly, as opposed to elementwise on arrays which is wrapped into a UDF) to a
    single input column and returns a single output column. Caters for the case, where
    the input column is a nested array, in which case the function is applied to the
    innermost array.

    `func` should be provided as a python function, and not wrapped in the Spark UDF
    wrapper.

    :param input_col: Input column.
    :param input_col_datatype: Input column datatype.
    :param func: Function to apply to the input column.
    :param udf_return_element_datatype: Datatype of the UDF return type. Should be the
    raw underlying type.
    :returns: Output column.
    """
    return _single_input_single_output_udf_transform(
        input_col, input_col_datatype, func, udf_return_element_datatype, scalar=False
    )


def _multi_input_single_output_transform(
    input_cols: List[Column],
    input_col_datatypes: List[DataType],
    input_col_names: List[str],
    func: Callable[[Column], Column],
    scalar: bool,
) -> Column:
    """
    Applies either a scalar or array Spark function to multiple input columns and
    returns a single output column.

    NOTE: This function expects a `func` that takes a single column returning a single
    column. Under the hood we will zip the input columns into a single column and apply
    `func` to the zipped column. Therefore `func` should be cognizant of the names of
    the elements in the zipped array column.

    Example adding two input columns together with names `input_0` and `input_1`:
    func = lambda x: x["input_0"] + x["input_1"]

    Ideally this function should not be used directly, the user
    should choose one of the below scalar vs array functions for their usecase. As every
    transform is either scalar or array, not both. This function is used to reduce code
    duplication between the scalar and array functions.

    :param input_cols: List of input columns.
    :param input_col_datatypes: List of input column datatypes.
    :param input_col_names: List of input column names. If any of the columns are
    arrays, the names of the elements in the zipped array column will be the same as
    these names.
    :param func: Function to apply to the zipped input column.
    :param scalar: Whether the function to be applied is a scalar function or an
    array function.
    :returns: Output column.
    """
    inputs_are_scalar = [
        not isinstance(datatype, ArrayType) for datatype in input_col_datatypes
    ]
    inputs_are_array = [
        isinstance(datatype, ArrayType) for datatype in input_col_datatypes
    ]

    if not scalar and any(inputs_are_scalar):
        raise ValueError(
            f"""Expected all input columns to be of type ArrayType,
            received {input_col_datatypes} instead."""
        )

    if not (all(inputs_are_array) or all(inputs_are_scalar)):
        # If the inputs are not either all scalar or all arrays, then we have a mix of
        # scalars and arrays. In this case we need to broadcast the scalars to the size
        # of the arrays.
        scalar_columns = [
            (idx, col_w_datatype[0])
            for idx, col_w_datatype in enumerate(zip(input_cols, input_col_datatypes))
            if not isinstance(col_w_datatype[1], ArrayType)
        ]
        array_columns_w_types = [
            (idx, col_w_datatype[0], col_w_datatype[1])
            for idx, col_w_datatype in enumerate(zip(input_cols, input_col_datatypes))
            if isinstance(col_w_datatype[1], ArrayType)
        ]
        # Broadcast the scalar to the size of the arrays. Use the first array column
        # to determine the size of the broadcasted scalar. Assumes all arrays are
        # of the same size.
        broadcasted_scalars = [
            (
                idx,
                broadcast_scalar_column_to_array(
                    scalar_column=scalar_column,
                    array_column=array_columns_w_types[0][1],
                    array_column_datatype=array_columns_w_types[0][2],
                ),
            )
            for idx, scalar_column in scalar_columns
        ]
        # Resort the array columns and the broadcasted scalars, so they match the order
        # of the input columns.
        columns_w_idx = [
            (idx, column) for idx, column, _ in array_columns_w_types
        ] + broadcasted_scalars
        sorted_columns = [
            column for idx, column in sorted(columns_w_idx, key=lambda x: x[0])
        ]
        # Check all arrays have same nesting level
        nested_levels = [
            get_array_nesting_level(column_dtype)
            for _, _, column_dtype in array_columns_w_types
        ]
        if not all(nested_level == nested_levels[0] for nested_level in nested_levels):
            raise ValueError(
                f"""Expected all input columns to have the same array nesting level,
                received {nested_levels} instead."""
            )
        nested_level = nested_levels[0]
        zipped_array_column = nested_arrays_zip(
            columns=sorted_columns,
            nest_level=nested_level,
            column_names=input_col_names,
        )
    else:
        # Otherwise, we have all scalars or all arrays and so can just zip together
        # the input columns.
        nested_level = get_array_nesting_level(column_dtype=input_col_datatypes[0])
        zipped_array_column = nested_arrays_zip(
            columns=input_cols, nest_level=nested_level, column_names=input_col_names
        )

    if not scalar:
        # If the function is not scalar then it operates on the full array, therefore
        # we reduce the nested level by 1.
        nested_level -= 1

    nested_func = nested_transform(func=func, nest_level=nested_level)
    return nested_func(zipped_array_column)


def multi_input_single_output_scalar_transform(
    input_cols: List[Column],
    input_col_datatypes: List[DataType],
    input_col_names: List[str],
    func: Callable[[Column], Column],
) -> Column:
    """
    Applies a scalar Spark function (e.g. a Spark standard library function that can be
    applied elementwise to arrays) to multiple input columns and returns a single output
    column. Caters for the case, where the input columns are:

    1. All scalars.
    2. All (possibly nested) arrays.
    3. A mix of scalars and (possibly nested) arrays.

    Zips the arrays into a single (potentially nested array) column.
    Applies `func` to the zipped column. `func` must be cognizant of the names of the
    elements in the zipped array column. If the inputs are a mix of scalars and arrays,
    broadcast the scalar to the size of the arrays before zipping the input columns.

    Example of `func` for a transformer that sums a list of input columns:

    `func = lambda x: reduce(add, [x[c] for c in input_col_names])`

    :param input_cols: List of input columns.
    :param input_col_datatypes: List of input column datatypes.
    :param input_col_names: List of input column names. If any of the columns are
    arrays, the names of the elements in the zipped array column will be the same as
    these names.
    :param func: Function to apply to the input columns in the case they are all scalar.
    :returns: Output column.
    """
    return _multi_input_single_output_transform(
        input_cols=input_cols,
        input_col_datatypes=input_col_datatypes,
        input_col_names=input_col_names,
        func=func,
        scalar=True,
    )


def multi_input_single_output_array_transform(
    input_cols: List[Column],
    input_col_datatypes: List[DataType],
    input_col_names: List[str],
    func: Callable[[Column], Column],
) -> Column:
    """
    Applies an array Spark function (e.g. a Spark standard library function that
    operates on an array directly, as opposed to elementwise on arrays) to multiple
    input columns and returns a single output column. Caters for the case, where the
    input column is a nested array, in which case the function is applied to the
    innermost array.

    NOTE: Function zips the multple inputs into a single column and so `func` should
    reference the input names of the new zipped struct column.

    Example:
    `func = lambda x: reduce(add, [x[c] for c in input_col_names])`

    :param input_cols: List of input columns.
    :param input_col_datatypes: List of input column datatypes.
    :param input_col_names: List of input column names. If any of the columns are
    arrays, the names of the elements in the zipped array column will be the same as
    these names.
    :param func: Function to apply to the input columns in the case they are all scalar.
    arrays. Main difference here is that the function must be cognizant of the names of
    the elements in the zipped array column.
    :returns: Output column.
    """
    return _multi_input_single_output_transform(
        input_cols=input_cols,
        input_col_datatypes=input_col_datatypes,
        input_col_names=input_col_names,
        func=func,
        scalar=False,
    )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

import numpy as np
import pyspark.sql.functions as F
from pyspark import keyword_only
from pyspark.sql import DataFrame
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import MaskValueParams, SingleInputSingleOutputParams
from kamae.spark.transformers import StandardScaleTransformer
from kamae.spark.utils import flatten_nested_arrays

from .base import BaseEstimator


class SingleFeatureArrayStandardScaleEstimator(
    BaseEstimator,
    SingleInputSingleOutputParams,
    MaskValueParams,
):
    """
    Single feature array standard scaler estimator for use in Spark pipelines.
    This estimator is used to calculate the mean and standard deviation of the input
    feature column when it is an array where all the elements represent the same
    feature. An example would be a sequence of trip durations or booking windows in
    a traveller's session. When fit is called it returns a StandardScaleTransformer
    which can be used to standardize/transform additional features, where the mean
    and standard deviation are calculated across all elements in all the arrays.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        maskValue: Optional[float] = None,
    ) -> None:
        """
        Initializes a SingleFeatureArrayStandardScaleEstimator estimator.
        Sets all parameters to given inputs.

        :param inputCol: Input column name to standardize.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
         in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(maskValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [FloatType(), DoubleType()]

    def _fit(self, dataset: DataFrame) -> "StandardScaleTransformer":
        """
        Fits the SingleFeatureArrayStandardScaleEstimator estimator to the given
        dataset. Calculates the mean and standard deviation of the input feature column
        and returns a StandardScaleTransformer with the mean and standard deviation set.

        All rows are assumed to be of the same length. The mask value which is set in
        the estimator is used to ignore certain values in the process of calculating
        the mean and stddev.
        :param dataset: Pyspark dataframe to fit the estimator to.
        :returns: StandardScaleTransformer instance with mean & standard deviation set.
        """

        input_column_type = self.get_column_datatype(dataset, self.getInputCol())
        if not isinstance(input_column_type, ArrayType):
            raise ValueError(
                f"""Input column {self.getInputCol()} must be of ArrayType.
                        Got {input_column_type} instead."""
            )

        # Collect a single row to driver and get the length.
        # We assume all subsequent rows have the same length.
        array_size = np.array((dataset.select(self.getInputCol()).first()[0])).shape[-1]

        # Flatten the array to a single array.
        # Will do nothing if the array is not nested.
        flattened_array_col = flatten_nested_arrays(
            column=F.col(self.getInputCol()), column_data_type=input_column_type
        )

        mean_and_stddev_dict: Dict[str, float] = (
            dataset.select(F.explode(flattened_array_col).alias(self.getInputCol()))
            .withColumn(
                "mask",
                F.when(
                    F.col(self.getInputCol()) == F.lit(self.getMaskValue()), 1
                ).otherwise(0),
            )
            .filter(F.col("mask") == F.lit(0))
            .agg(
                F.mean(self.getInputCol()).alias("mean"),
                F.stddev_pop(self.getInputCol()).alias("stddev"),
            )
            .first()
            .asDict()
        )
        mean: List[float] = [mean_and_stddev_dict["mean"] for _ in range(array_size)]
        stddev: List[float] = [
            mean_and_stddev_dict["stddev"] for _ in range(array_size)
        ]

        return StandardScaleTransformer(
            inputCol=self.getInputCol(),
            outputCol=self.getOutputCol(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            mean=mean,
            stddev=stddev,
            maskValue=self.getMaskValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Union

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class MultiplyLayer(BaseLayer):
    """
    Performs the multiply(x, y) operation on a given input tensor.
    If multiplier is not set, inputs are assumed to be a list of tensors and multiplied.
    If multiplier is set, inputs must be a tensor.
    """

    def __init__(
        self,
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        multiplier: float = None,
        **kwargs,
    ) -> None:
        """
        Initializes the MultiplyLayer layer

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param multiplier: The multiplier to multiply the input by, defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.multiplier = multiplier

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.uint8,
            tf.int8,
            tf.uint16,
            tf.int16,
            tf.int32,
            tf.int64,
            tf.complex64,
            tf.complex128,
        ]

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Union[Tensor, Iterable[Tensor]], **kwargs) -> Tensor:
        """
        Performs the multiply(x, y) operation on either an iterable of input tensors or
        a single input tensor and a constant.

        Decorated with `@allow_single_or_multiple_tensor_input` to ensure that the input
        is either a single tensor or an iterable of tensors. Returns this result as a
        list of tensors for easier use here.

        :param inputs: Single tensor or iterable of tensors to perform the
        multiply(x, y) operation on.
        :returns: The tensor resulting from the multiply(x, y) operation.
        """
        if self.multiplier is not None:
            if len(inputs) > 1:
                raise ValueError("If multiplier is set, cannot have multiple inputs")
            cast_input, cast_multiplier = self._force_cast_to_compatible_numeric_type(
                inputs[0], self.multiplier
            )
            return tf.math.multiply(
                cast_input,
                cast_multiplier,
            )
        else:
            if not len(inputs) > 1:
                raise ValueError("If multiplier is not set, must have multiple inputs")

            return reduce(tf.math.multiply, inputs)

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the Multiply layer.
        Used for saving and loading from a model.

        Specifically adds the `multiplier` to the config dictionary.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"multiplier": self.multiplier})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import enforce_single_tensor_input

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class RoundLayer(BaseLayer):
    """
    Performs a standard rounding operation on the input tensor.
    Supported rounding types are 'ceil', 'floor' and 'round'.

    - 'ceil' rounds up to the nearest integer.
    - 'floor' rounds down to the nearest integer.
    - 'round' rounds to the nearest integer.
    """

    def __init__(
        self,
        round_type: str = "round",
        name: str = None,
        input_dtype: str = None,
        output_dtype: str = None,
        **kwargs,
    ) -> None:
        """
        Initialises the RoundLayer layer.

        :param round_type: The type of rounding to perform.
        Supported types are 'ceil', 'floor' and 'round'. Defaults to 'round'.
        :param name: The name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        if round_type not in ["ceil", "floor", "round"]:
            raise ValueError("""roundType must be one of 'ceil', 'floor' or 'round'.""")
        self.round_type = round_type

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [tf.float16, tf.float32, tf.float64]

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs the rounding operation on the input tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that the input is a
        single tensor. Raises an error if multiple tensors are passed in as an iterable.

        :param inputs: Input tensor to perform the rounding on.
        :returns: The input tensor with the rounding applied.
        """
        if self.round_type == "ceil":
            return tf.math.ceil(inputs)
        elif self.round_type == "floor":
            return tf.math.floor(inputs)
        elif self.round_type == "round":
            return tf.math.round(inputs)
        else:
            raise ValueError("""roundType must be one of 'ceil', 'floor' or 'round'.""")

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the Round layer.
        Used for saving and loading from a model.

        Specifically adds the `round_type` value to the configuration.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update({"round_type": self.round_type})
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import numpy as np
import pyspark.sql.functions as F
from numpy.typing import NDArray
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import ArrayType, DataType, DoubleType, FloatType

from kamae.spark.params import (
    NanFillValueParams,
    SingleInputSingleOutputParams,
    StandardScaleSkipZerosParams,
)
from kamae.spark.transformers import ConditionalStandardScaleTransformer
from kamae.spark.utils import construct_nested_elements_for_scaling
from kamae.utils import get_condition_operator

from .base import BaseEstimator


class ConditionalStandardScaleEstimatorParams(Params):
    """
    Mixin class containing conditional standard scale parameters,
    needed for single feature array scaler layers.
    """

    scalingFunction = Param(
        Params._dummy(),
        "scalingFunction",
        """
        The name of the scaling function to use during spark fit function,
        to estimate the mean and standard deviation. Defaults to 'standard'.
        """,
        typeConverter=TypeConverters.toString,
    )

    maskCols = Param(
        Params._dummy(),
        "maskCols",
        """
        Columns on which to apply the mask condition and the mask value
        during moments calculation. Defaults to None.
        """,
        typeConverter=TypeConverters.toListString,
    )

    maskOperators = Param(
        Params._dummy(),
        "maskOperators",
        "Operators to use in masking conditions: eq, neq, lt, gt, leq, geq",
        typeConverter=TypeConverters.toListString,
    )

    maskValues = Param(
        Params._dummy(),
        "maskValues",
        """
        Values applied to the respective maskCol with the respective maskOperator
        to make the value ignored during moments calculation.
        """,
        typeConverter=TypeConverters.toListFloat,
    )

    relevanceCol = Param(
        Params._dummy(),
        "relevanceCol",
        """
        The name of the relevance column to use during the calculation of the moments.
        """,
        typeConverter=TypeConverters.toString,
    )

    def setScalingFunction(
        self,
        value: str,
    ) -> "ConditionalStandardScaleEstimatorParams":
        """
        Sets the scalingFunction parameter.

        :param value: String value to indicate which scaling function to use.
        :returns: Instance of class mixed in.
        """
        if value.lower() == "standard":
            return self._set(scalingFunction=value.lower())
        elif value.lower() == "binary":
            return self._set(scalingFunction=value.lower())
        else:
            raise ValueError(f"Unknown scaling function: {value}.")

    def getScalingFunction(self) -> str:
        """
        Gets the scalingFunction parameter.

        :returns: Boolean value of the scalingFunction value.
        """
        return self.getOrDefault(self.scalingFunction)

    def setMaskCols(
        self, value: List[str]
    ) -> "ConditionalStandardScaleEstimatorParams":
        """
        Sets the maskCols parameter.

        :param value: Columns to use as the mask columns.
        :returns: Instance of class mixed in.
        """
        return self._set(maskCols=value)

    def getMaskCols(self) -> List[str]:
        """
        Gets the maskCols parameter.
        :returns: List of string values of the mask column.
        """
        return self.getOrDefault(self.maskCols)

    def setMaskOperators(
        self, value: List[str]
    ) -> "ConditionalStandardScaleEstimatorParams":
        """
        Sets the maskOperators parameter.

        :param value: String value describing the operator to use in condition:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        :returns: Instance of class mixed in.
        """
        allowed_operators = ["eq", "neq", "lt", "gt", "leq", "geq"]
        for v in value:
            if v not in allowed_operators:
                raise ValueError(
                    f"conditionOperator must be one of {allowed_operators}, "
                    f"but got {value}"
                )
        return self._set(maskOperators=value)

    def getMaskOperators(self) -> List[str]:
        """
        Gets the maskOperators parameter.

        :returns: List of string values describing the operators to use in conditions:
        - eq
        - neq
        - lt
        - gt
        - leq
        - geq
        """
        return self.getOrDefault(self.maskOperators)

    def setMaskValues(
        self, value: List[float]
    ) -> "ConditionalStandardScaleEstimatorParams":
        """
        Sets the maskValues parameter.

        :param value: List of float values to use as the mask value.
        :returns: Instance of class mixed in.
        """
        return self._set(maskValues=value)

    def getMaskValues(self) -> List[float]:
        """
        Gets the maskValues parameter.

        :returns: List of float values of the mask values.
        """
        return self.getOrDefault(self.maskValues)

    def setRelevanceCol(self, value: str) -> "ConditionalStandardScaleEstimatorParams":
        """
        Sets the relevanceCol parameter.

        :param value: String value to use as the relevance column.
        :returns: Instance of class mixed in.
        """
        return self._set(relevanceCol=value)

    def getRelevanceCol(self) -> str:
        """
        Gets the relevanceCol parameter.

        :returns: String value of the relevance column.
        """
        return self.getOrDefault(self.relevanceCol)


class ConditionalStandardScaleEstimator(
    BaseEstimator,
    SingleInputSingleOutputParams,
    ConditionalStandardScaleEstimatorParams,
    StandardScaleSkipZerosParams,
    NanFillValueParams,
):
    """
    Conditional standard scaler estimator for use in Spark pipelines.
    This is used to calculate the mean and standard deviation with masking,
    and then to standardize/transform the input column using the mean and
    standard deviation, optionally skipping the standardization of inputs
    equal to zero.
    The mask columns, mask values and mask conditions are used to calculate
    the moments only when all the masking conditions are satisfied.
    The skip_zeros parameter allows to apply both the scaling and the transformation
    only when input is not equal to zero. If equal to zero, it will remain zero in
    the output value as it was in the input value.
    It is also possible to specify a non-standard scaling function using the
    scalingFunction parameter.
    When fit is called it returns a ConditionalStandardScaleTransformer
    which can be used to standardize/transform the input data.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        layerName: Optional[str] = None,
        scalingFunction: str = "standard",
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        maskCols: Optional[List[str]] = None,
        maskOperators: Optional[List[str]] = None,
        maskValues: Optional[List[float]] = None,
        relevanceCol: Optional[str] = None,
        skipZeros: bool = False,
        epsilon: float = 0,
        nanFillValue: Optional[float] = None,
    ) -> None:
        """
        Initializes a ConditionalStandardScaleEstimator estimator.
        Sets all parameters to given inputs.

        :param inputCol: Input column name to standardize.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param scalingFunction: The scaling function to use: 'standard', 'binary'.
        :param maskCols: Columns on which to apply the mask values. Defaults to None.
        :param maskOperators: Operators to use in each masking condition:
        eq, neq, lt, gt, leq, geq. Defaults to 'neq'.
        :param maskValues: Values applied to the maskCols which makes the value
        ignored in the calculation of the moments. Defaults to None.
        :param relevanceCol: The name of the relevance column to use during spark
        fit function.
        :param skipZeros: If True, during spark transform and keras inference,
        do not apply the scaling when the values to scale are equal to zero.
        :param epsilon: Small value to add to conditional check of zeros. Valid only
        when skipZeros is True. Defaults to 0.
        :param nanFillValue: Value to fill NaNs with after scaling. It is important
        to use it if epsilon filters out all the values. Defaults to None.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(
            scalingFunction="standard",
            maskCols=None,
            maskOperators=None,
            maskValues=None,
            relevanceCol=None,
            skipZeros=False,
            epsilon=0,
            nanFillValue=None,
        )
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        Returns the compatible data types for transformer.
        :returns: List of compatible data types.
        """
        return [DoubleType(), FloatType()]

    def _get_binary_moments(self, f: NDArray, n: NDArray) -> (NDArray, NDArray):
        """
        Calculates the moments for a binary variable.

        :param f: Number of samples where feature is one/true.
        :param n: Number of samples.
        :returns:
            - mean - The mean of the distribution.
            - stddev - The stddev of the distribution.
        """
        true_ratio = np.where(n <= 0, 0, f / n)
        variance = np.where(
            (n - 1) <= 0,
            0,
            (f * pow(1 - true_ratio, 2) + (n - f) * pow(0 - true_ratio, 2)) / (n - 1),
        )
        mean = 1 - true_ratio
        stddev = np.sqrt(variance)
        return mean, stddev

    def _validate_mask_ops(self) -> None:
        """
        Validates the mask operators.
        Mask columns, operators and values must be set together
        and must have the same length.
        """
        arr1 = self.getMaskCols()
        arr2 = self.getMaskOperators()
        arr3 = self.getMaskValues()
        if any(arr is not None for arr in [arr1, arr2, arr3]):
            if any(arr is None or len(arr) != len(arr1) for arr in [arr1, arr2, arr3]):
                raise ValueError(
                    "Mask columns, operators and values must be set together "
                    "and must have the same length."
                )
        return

    def _fit(self, dataset: DataFrame) -> "ConditionalStandardScaleTransformer":
        """
        Fits the ConditionalStandardScaleEstimator estimator to the given dataset.
        Calculates the mean and standard deviation of the input feature column and
        returns a StandardScaleTransformer with the mean and standard deviation set.

        :param dataset: Pyspark dataframe to fit the estimator to.
        :returns: ConditionalStandardScaleEstimator instance with
        mean & standard deviation set.
        """
        input_column_dtype = self.get_column_datatype(dataset, self.getInputCol())
        if not isinstance(input_column_dtype, ArrayType):
            input_col = F.array(F.col(self.getInputCol()))
            input_column_dtype = ArrayType(input_column_dtype)
        else:
            input_col = F.col(self.getInputCol())

        # Masks are applied to the dataset before calculating the moments
        self._validate_mask_ops()
        if self.getMaskCols() is not None:
            mask_cols = self.getMaskCols()
            for i in range(len(mask_cols)):
                mask_col = mask_cols[i]
                if mask_col not in dataset.columns:
                    raise ValueError(f"Mask column {mask_col} not found in dataset.")
                mask_op = get_condition_operator(self.getMaskOperators()[i])
                mask_val = self.getMaskValues()[i]
                dataset = dataset.filter(mask_op(F.col(mask_col), mask_val))

        # Collect a single row to driver and get the length.
        # We assume all subsequent rows have the same length.
        row = dataset.select(input_col).first()
        if row is None:
            raise ValueError("No data left after application of mask conditions.")
        array_size = np.array((row[0])).shape[-1]

        # Calculate the moments
        if self.getScalingFunction().lower() == "standard":
            return self._fit_standard(
                dataset, input_col, input_column_dtype, array_size
            )
        elif self.getScalingFunction().lower() == "binary":
            return self._fit_binary(dataset, input_col, input_column_dtype, array_size)
        else:
            raise ValueError(f"Unknown scaling function: {self.getScalingFunction()}.")

    def _fit_binary(
        self,
        dataset: DataFrame,
        input_column: Column,
        input_column_dtype: DataType,
        array_size: int,
    ) -> "ConditionalStandardScaleTransformer":
        """
        Fits the ConditionalStandardScaleEstimator estimator with
        the binary scaling function. In this case, the relevance
        column name must be set. This should be used only when input
        is a binary variable and the label can be transformed into a
        classification task (if relevance > 0 then 1 else 0).

        With this function, the mean and stddev are:
            mean = 1 - (f/n)
            stddev = sqrt((f * pow(1-(f/n), 2) + (n-f) * pow(0-(f/n), 2)) / (n-1))
        where:
            n = sum(when(x==1, 1, 0))
            f = sum(when(x==1 && relevance > 0, 1, 0))

        :param dataset: Pyspark dataframe to fit the estimator to.
        :param array_size: The size of the array to standardize.
        :returns: ConditionalStandardScaleEstimator instance with
        mean & standard deviation set.
        """
        if self.getRelevanceCol() is None:
            raise ValueError("Relevance column must be set for binary scaling.")

        # Construct the elements to calculate the moments
        element_struct = construct_nested_elements_for_scaling(
            column=input_column,
            column_datatype=input_column_dtype,
            array_dim=array_size,
        )

        count_cols = [
            F.sum(
                F.when(
                    F.col(f"element_struct.element_{i}") == F.lit(1),
                    1,
                ).otherwise(0)
            ).alias(f"count_{i}")
            for i in range(1, array_size + 1)
        ]
        count_ones_cols = [
            F.sum(
                F.when(
                    (F.col(f"element_struct.element_{i}") == F.lit(1))
                    & (F.col(self.getRelevanceCol()) > 0),
                    1,
                ).otherwise(0)
            ).alias(f"count_ones_{i}")
            for i in range(1, array_size + 1)
        ]

        # apply the aggregations
        metric_cols = count_cols + count_ones_cols
        metrics_dict = (
            dataset.withColumn("element_struct", element_struct)
            .agg(*metric_cols)
            .first()
            .asDict()
        )
        count = [metrics_dict[f"count_{i}"] for i in range(1, array_size + 1)]
        count_ones = [metrics_dict[f"count_ones_{i}"] for i in range(1, array_size + 1)]
        if self.getNanFillValue() is not None:
            fill_val = self.getNanFillValue()
            count = [fill_val if c is None else c for c in count]
            count_ones = [fill_val if c is None else c for c in count_ones]
        mean, stddev = self._get_binary_moments(np.array(count_ones), np.array(count))
        return ConditionalStandardScaleTransformer(
            inputCol=self.getInputCol(),
            outputCol=self.getOutputCol(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            mean=mean.tolist(),
            stddev=stddev.tolist(),
            skipZeros=self.getSkipZeros(),
            epsilon=self.getEpsilon(),
        )

    def _fit_standard(
        self,
        dataset: DataFrame,
        input_column: Column,
        input_column_dtype: DataType,
        array_size: int,
    ) -> "ConditionalStandardScaleTransformer":
        """
        Fits the ConditionalStandardScaleEstimator estimator with
        the standard scaling function.
        This should be the default function for the scaling operation.

        :param dataset: Pyspark dataframe to fit the estimator to.
        :param array_size: The size of the array to standardize.
        :returns: ConditionalStandardScaleEstimator instance with
        mean & standard deviation set.
        """
        # Construct the elements to calculate the moments
        element_struct = construct_nested_elements_for_scaling(
            column=input_column,
            column_datatype=input_column_dtype,
            array_dim=array_size,
        )
        # Defaults
        mean_cols = [
            F.mean(F.col(f"element_struct.element_{i}")).alias(f"mean_{i}")
            for i in range(1, array_size + 1)
        ]
        stddev_cols = [
            F.stddev_pop(F.col(f"element_struct.element_{i}")).alias(f"stddev_{i}")
            for i in range(1, array_size + 1)
        ]
        # Use relevance column and skip zeros (with epsilon)
        if self.getSkipZeros() and (self.getRelevanceCol() is not None):
            eps = self.getEpsilon()
            mean_cols = [
                F.mean(
                    F.when(
                        # x != (0 +- eps)
                        (F.abs(F.col(f"element_struct.element_{i}")) > F.lit(eps))
                        & (F.col(self.getRelevanceCol()) > 0),
                        F.col(f"element_struct.element_{i}"),
                    )
                ).alias(f"mean_{i}")
                for i in range(1, array_size + 1)
            ]
            stddev_cols = [
                F.stddev_pop(
                    F.when(
                        # x != (0 +- eps)
                        (F.abs(F.col(f"element_struct.element_{i}")) > F.lit(eps))
                        & (F.col(self.getRelevanceCol()) > 0),
                        F.col(f"element_struct.element_{i}"),
                    )
                ).alias(f"stddev_{i}")
                for i in range(1, array_size + 1)
            ]
        # Use relevance column
        elif self.getRelevanceCol() is not None:
            mean_cols = [
                F.mean(
                    F.when(
                        (F.col(self.getRelevanceCol()) > 0),
                        F.col(f"element_struct.element_{i}"),
                    )
                ).alias(f"mean_{i}")
                for i in range(1, array_size + 1)
            ]
            stddev_cols = [
                F.stddev_pop(
                    F.when(
                        (F.col(self.getRelevanceCol()) > 0),
                        F.col(f"element_struct.element_{i}"),
                    )
                ).alias(f"stddev_{i}")
                for i in range(1, array_size + 1)
            ]
        # Skip zeros on fit (with epsilon)
        elif self.getSkipZeros():
            eps = self.getEpsilon()
            mean_cols = [
                F.mean(
                    F.when(
                        # x != (0 +- eps)
                        F.abs(F.col(f"element_struct.element_{i}")) > F.lit(eps),
                        F.col(f"element_struct.element_{i}"),
                    ),
                ).alias(f"mean_{i}")
                for i in range(1, array_size + 1)
            ]
            stddev_cols = [
                F.stddev_pop(
                    F.when(
                        # x != (0 +- eps)
                        F.abs(F.col(f"element_struct.element_{i}")) > F.lit(eps),
                        F.col(f"element_struct.element_{i}"),
                    ),
                ).alias(f"stddev_{i}")
                for i in range(1, array_size + 1)
            ]
        # apply the aggregations
        metric_cols = mean_cols + stddev_cols
        mean_and_stddev_dict = (
            dataset.withColumn("element_struct", element_struct)
            .agg(*metric_cols)
            .first()
            .asDict()
        )
        mean = [mean_and_stddev_dict[f"mean_{i}"] for i in range(1, array_size + 1)]
        stddev = [mean_and_stddev_dict[f"stddev_{i}"] for i in range(1, array_size + 1)]
        if self.getNanFillValue() is not None:
            fill_val = self.getNanFillValue()
            mean = [fill_val if m is None else m for m in mean]
            stddev = [fill_val if s is None else s for s in stddev]
        return ConditionalStandardScaleTransformer(
            inputCol=self.getInputCol(),
            outputCol=self.getOutputCol(),
            layerName=self.getLayerName(),
            inputDtype=self.getInputDtype(),
            outputDtype=self.getOutputDtype(),
            mean=mean,
            stddev=stddev,
            skipZeros=self.getSkipZeros(),
            epsilon=self.getEpsilon(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import SingleInputSingleOutputParams, UnixTimestampParams
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import DateTimeToUnixTimestampLayer


class DateTimeToUnixTimestampTransformer(
    BaseTransformer, SingleInputSingleOutputParams, UnixTimestampParams
):
    """
    Transformer that converts a datetime string to a unix timestamp.

    The unix timestamp can be in milliseconds or seconds, set by the `unit` parameter.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        unit: str = "s",
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initialises the DateTimeToUnixTimestampTransformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param unit: Unit of the output timestamp. Can be `milliseconds`
        (shorthand `ms`) or `seconds` (shorthand `s`). Default is `s` (seconds).
        :param layerName: Layer name. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(unit="s")
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            StringType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input integer timestamp to the date string with format
        yyyy-MM-dd HH:mm:ss.SSS.

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def datetime_to_unix_timestamp(datetime: Column) -> Column:
            """
            Converts a datetime string to a unix timestamp in seconds.

            :param datetime: Column of datetime strings.
            :returns: Column of unix timestamps.
            """
            # Check if we have a date only or datetime. This is quite a crude way to
            # check if we have a datetime string.
            split_datetime = F.split(datetime, " ")
            is_datetime = F.size(split_datetime) > 1
            # If we have a date only, add 00:00:00.000 UTC. Otherwise, add UTC suffix
            # This is to ensure that the datetime string is in the correct format for
            # PySpark
            datetime_w_utc_tz = F.when(
                is_datetime, F.concat(datetime, F.lit(" UTC"))
            ).otherwise(F.concat(datetime, F.lit(" 00:00:00.000 UTC")))
            # Convert datetime string to timestamp
            datetime_timestamp = F.to_timestamp(datetime_w_utc_tz)
            # Convert timestamp to unix timestamp
            unix_timestamp_wo_milliseconds = F.unix_timestamp(datetime_timestamp)
            # Extract milliseconds from the datetime string
            milliseconds_str = F.date_format(datetime_timestamp, "SSS")
            # Convert milliseconds to float
            milliseconds_float = milliseconds_str.cast("float") / 1000.0
            # Add milliseconds to the unix timestamp if we have them
            return unix_timestamp_wo_milliseconds + milliseconds_float

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: datetime_to_unix_timestamp(x)
            if self.getUnit() == "s"
            else datetime_to_unix_timestamp(x) * 1000.0,
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer that performs the datetime to unix timestamp.

        :returns: Tensorflow layer that performs the unix timestamp to date transform.
        """
        return DateTimeToUnixTimestampLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            unit=self.getUnit(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import allow_single_or_multiple_tensor_input, get_top_n

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ListStdDevLayer(BaseLayer):
    """
    Calculate the average across the axis dimension.
    - If one tensor is passed, the transformer calculates the average of the tensor
    based on all the items in the given axis dimension.
    - If inputCols is set, the transformer calculates the average of the first tensor
    based on second tensor's topN items in the same given axis dimension.

    By using the topN items to calculate the statistics, we can better approximate
    the real statistics in production. It is suggested to use a large enough topN to
    get a good approximation of the statistics, and an important feature to sort on,
    such as item's past production.

    Example: calculate the average price in the same query, based only on the top N
    items sorted by descending production.
    """

    def __init__(
        self,
        name: str,
        input_dtype: str = None,
        output_dtype: str = None,
        top_n: int = None,
        sort_order: str = "asc",
        min_filter_value: float = None,
        nan_fill_value: float = 0.0,
        axis: int = 1,
        **kwargs,
    ):
        """
        Initializes the Listwise Average layer.

        WARNING: The code is fully tested for axis=1 only. Further testing is needed.

        WARNING: The code can be affected by the value of the padding items. Always
        make sure to filter out the padding items value with min_filter_value.

        :param name: Name of the layer, defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param top_n: The number of top items to consider when calculating the average.
        :param sort_order: The order to sort the second tensor by. Defaults to `asc`.
        :param min_filter_value: The minimum filter value to ignore values during
        calculation. Defaults to None (no filter).
        :param nan_fill_value: The value to fill NaNs results with. Defaults to 0.
        :param axis: The axis to calculate the statistics across. Defaults to 1.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )
        self.top_n = top_n
        self.sort_order = sort_order
        self.min_filter_value = min_filter_value
        self.nan_fill_value = nan_fill_value
        self.axis = axis

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer.

        :returns: The compatible dtypes of the layer.
        """
        return [
            tf.bfloat16,
            tf.float16,
            tf.float32,
            tf.float64,
            tf.uint8,
            tf.int8,
            tf.uint16,
            tf.int16,
            tf.int32,
            tf.int64,
            tf.complex64,
            tf.complex128,
        ]

    @allow_single_or_multiple_tensor_input
    def _call(self, inputs: Iterable[Tensor], **kwargs) -> Tensor:
        """
        Calculate the listwise average, optionally sorting and
        filtering based on the second input tensor.

        :param inputs: The iterable tensor for the feature.
        :returns: Thew new tensor result column.
        """
        val_tensor = inputs[0]
        output_shape = tf.shape(val_tensor)

        with_sort = True if len(inputs) == 2 else False
        sort_tensor = inputs[1] if with_sort else None

        if with_sort and self.top_n is None:
            raise ValueError("topN must be specified when using a sort column.")

        if with_sort:
            # Get the values corresponding to the top N item in the sort tensor
            filtered_tensor = get_top_n(
                val_tensor=val_tensor,
                axis=self.axis,
                sort_tensor=sort_tensor,
                sort_order=self.sort_order,
                top_n=self.top_n,
            )
        else:
            filtered_tensor = val_tensor

        # Apply the mask to filter out elements less than or equal to the threshold
        if self.min_filter_value is not None:
            mask = tf.greater_equal(filtered_tensor, self.min_filter_value)
            nan_tensor = tf.constant(float("nan"), dtype=val_tensor.dtype)
            filtered_tensor = tf.where(mask, filtered_tensor, nan_tensor)
            mask = tf.math.is_finite(filtered_tensor)
            numerator = tf.reduce_sum(
                tf.where(mask, filtered_tensor, tf.zeros_like(filtered_tensor)),
                axis=self.axis,
                keepdims=True,
            )
            denominator = tf.reduce_sum(
                tf.cast(mask, dtype=numerator.dtype),
                axis=self.axis,
                keepdims=True,
            )
            listwise_mean = tf.truediv(numerator, denominator)

        else:
            # Calculate the mean without filtering
            listwise_mean = tf.reduce_mean(
                filtered_tensor,
                axis=self.axis,
                keepdims=True,
            )

        # Calculate the squared differences from the mean
        squared_diff = tf.square(filtered_tensor - listwise_mean)

        # Calculate the sample variance by dividing the sum of squared diff by (N - 1)
        mask = tf.math.is_finite(squared_diff)
        listwise_sum = tf.reduce_sum(
            tf.where(mask, squared_diff, tf.zeros_like(squared_diff)),
            axis=self.axis,
            keepdims=True,
        )
        listwise_count = tf.reduce_sum(
            tf.cast(mask, dtype=listwise_sum.dtype),
            axis=self.axis,
            keepdims=True,
        )
        listwise_variance = tf.math.divide_no_nan(listwise_sum, (listwise_count - 1))
        listwise_stddev = tf.sqrt(listwise_variance)

        # Fill nan
        is_integer = listwise_stddev.dtype.is_integer
        nan_val = int(self.nan_fill_value) if is_integer else self.nan_fill_value
        listwise_stddev = tf.where(
            tf.math.is_nan(listwise_stddev),
            tf.constant(nan_val, dtype=listwise_mean.dtype),
            listwise_stddev,
        )

        # Broadcast the stat to each item in the list
        # WARNING: If filter creates empty items list, the result will be NaN
        listwise_stddev = tf.broadcast_to(listwise_stddev, output_shape)

        return listwise_stddev

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "top_n": self.top_n,
                "sort_order": self.sort_order,
                "min_filter_value": self.min_filter_value,
                "nan_fill_value": self.nan_fill_value,
                "axis": self.axis,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.sql import Column, DataFrame, SparkSession
from pyspark.sql.types import DataType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.transformers.base import BaseTransformer
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import CurrentDateLayer


class CurrentDateTransformer(BaseTransformer, SingleInputSingleOutputParams):
    """
    Returns the current UTC date in yyyy-MM-dd format.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
    ) -> None:
        """
        Initialises the CurrentDateTransformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :returns: None - class instantiated.
        """
        super().__init__()
        kwargs = self._input_kwargs
        self.setParams(**kwargs)
        # TODO: Remove this when we only support PySpark 3.5+. It is only used to get
        #  the timezone set by the user for datetime operations. In 3.5+ we can use the
        #  current_timezone() function. Also is there a better way to access this than
        #  inside a class attribute? Setting it at the top of the file causes issues
        #  in tests as we import all transformers when the package is loaded.
        self.spark = SparkSession.builder.getOrCreate()

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return None

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Returns a column of the current date. If an array column is provided,
        we return an array column of identical structure with elements populated by
        the current date.

        :param dataset: Input dataframe.
        :returns: Transformed dataframe.
        """

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )

        def current_utc_date() -> Column:
            """
            Returns the current UTC date. Spark respects the timezone set in the Spark
            session so we need to convert the local timestamp to UTC before extracting
            the date.

            :returns: Column of the current UTC date.
            """
            local_timestamp = F.localtimestamp()
            # TODO: Replace this with current_timezone() once we only support PySpark
            #  3.5+
            local_timezone = self.spark.conf.get("spark.sql.session.timeZone")
            return F.to_date(F.to_utc_timestamp(local_timestamp, local_timezone))

        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: current_utc_date().cast("string"),
        )

        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer.

        :returns: CurrentDateLayer Tensorflow layer.
        """
        return CurrentDateLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
import re
from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import DataFrame
from pyspark.sql.types import DataType, StringType

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_scalar_transform
from kamae.tensorflow.layers import SubStringDelimAtIndexLayer

from .base import BaseTransformer


class SubStringDelimAtIndexParams(Params):
    """
    Mixin class containing delimiter & index parameter needed for sub string transforms.
    """

    delimiter = Param(
        Params._dummy(),
        "delimiter",
        "Value to use to split the string into substrings.",
        typeConverter=TypeConverters.toString,
    )

    index = Param(
        Params._dummy(),
        "index",
        "Once the string is split using delimiter, which index to return.",
        typeConverter=TypeConverters.toInt,
    )

    defaultValue = Param(
        Params._dummy(),
        "defaultValue",
        "If the index is out of bounds after string split, what value to return.",
        typeConverter=TypeConverters.toString,
    )

    def setDelimiter(self, value: str) -> "SubStringDelimAtIndexParams":
        """
        Sets the delimiter parameter.

        :param value: String value to split substring on.
        :returns: Instance of class mixed in.
        """
        return self._set(delimiter=value)

    def getDelimiter(self) -> str:
        """
        Gets the delimiter parameter.

        :returns: String value to split substring on.
        """
        return self.getOrDefault(self.delimiter)

    def setIndex(self, value: int) -> "SubStringDelimAtIndexParams":
        """
        Sets the delimiter parameter.

        :param value: Index of substring to return.
        :returns: Instance of class mixed in.
        """
        return self._set(index=value)

    def getIndex(self) -> int:
        """
        Gets the index parameter.

        :returns: Integer value of index of substring to return.
        """
        return self.getOrDefault(self.index)

    def setDefaultValue(self, value: str) -> "SubStringDelimAtIndexParams":
        """
        Sets the defaultValue parameter.

        :param value: String value use as default if index is out of bounds.
        :returns: Instance of class mixed in.
        """
        return self._set(defaultValue=value)

    def getDefaultValue(self) -> str:
        """
        Gets the defaultValue parameter.

        :returns: String value use as default if index is out of bounds.
        """
        return self.getOrDefault(self.defaultValue)


class SubStringDelimAtIndexTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    SubStringDelimAtIndexParams,
):
    """
    Sub string at delimiter Spark Transformer for use in Spark pipelines.
    This transformer splits a string at a delimiter and returns the substring
    at the specified index. If the delimiter is the empty string, the string
    is split by characters.
    If the index is negative, start counting from the end of the string.
    If the index is out of bounds, the default value is returned.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        delimiter: Optional[str] = None,
        index: Optional[int] = None,
        defaultValue: Optional[str] = None,
    ) -> None:
        """
        Initializes an SubStringDelimAtIndexTransformer transformer.

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        in the keras model. If not set, we use the uid of the Spark transformer.
        :param delimiter: Value to use to split the string into substrings.
        Default is "_".
        :param index: Once the string is split using delimiter, which index to return.
        If the index is negative, start counting from the end of the string.
        Default is 0.
        :param defaultValue: If the index is out of bounds after string split,
         what value to return. Default is empty string.
        :returns: None - class instantiated.
        """
        super().__init__()
        self._setDefault(delimiter="_")
        self._setDefault(defaultValue="")
        self._setDefault(index=0)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [StringType()]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Transforms the input dataset. Creates a new column with name `outputCol`,
        which splits the input column at the delimiter and returns the substring
        at the specified index. If the index is out of bounds, the default value
        is returned.

        :param dataset: Pyspark dataframe to transform.
        :returns: Transformed pyspark dataframe.
        """
        delimiter = re.escape(self.getDelimiter())
        index = self.getIndex()
        default_value = self.getDefaultValue()

        input_datatype = self.get_column_datatype(
            dataset=dataset, column_name=self.getInputCol()
        )
        # Since element_at is a 1-based index , we need to add 1 to the index if it
        # is non-negative.
        one_based_index = index + 1 if index >= 0 else index
        output_col = single_input_single_output_scalar_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_datatype,
            func=lambda x: F.coalesce(
                F.element_at(F.split(x, delimiter), one_based_index),
                F.lit(default_value),
            ),
        )
        return dataset.withColumn(self.getOutputCol(), output_col)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for SubStringDelimAtIndexTransformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
         performs sub string at delimiter.
        """
        return SubStringDelimAtIndexLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            delimiter=self.getDelimiter(),
            index=self.getIndex(),
            default_value=self.getDefaultValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import (
    enforce_single_tensor_input,
    unix_timestamp_to_datetime,
)

from .base import BaseLayer


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class CurrentDateTimeLayer(BaseLayer):
    """
    Returns the current timestamp in yyyy-MM-dd HH:mm:ss.SSS format.

    NOTE: Parity between this and its Spark counterpart is very difficult at the
    millisecond level. We have to round the TensorFlow timestamp to the 3rd decimal
    place for milliseconds, because  Spark already truncates to 3 decimal places.
    Therefore, parity is not guaranteed at this precision.

    It is recommended not to rely on parity at the millisecond level.
    """

    def __init__(
        self, name=None, input_dtype: str = None, output_dtype: str = None, **kwargs
    ) -> None:
        """
        Initialises an instance of the CurrentDateTimeLayer layer.

        :param name: Name of the layer. Defaults to `None`.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name, input_dtype=input_dtype, output_dtype=output_dtype, **kwargs
        )

    @property
    def compatible_dtypes(self) -> Optional[List[tf.dtypes.DType]]:
        """
        Returns the compatible dtypes of the layer. Returns `None` as the layer
        only returns the current date as a string. It does not transform any input.

        :returns: The compatible dtypes of the layer.
        """
        return None

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> tf.Tensor:
        """
        Returns the current timestamp in yyyy-MM-dd HH:mm:ss format.
        Uses the input tensor to determine the shape of the output tensor.

        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.

        :param inputs: Input tensor to determine the shape of the output tensor.
        :returns: The current timestamp tensor in yyyy-MM-dd format.
        """
        current_timestamp = tf.fill(tf.shape(inputs), tf.timestamp())
        outputs = unix_timestamp_to_datetime(current_timestamp, True)
        return outputs

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the CurrentDateTime layer.
        Used for saving and loading from a model.

        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import pyspark.sql.functions as F
import tensorflow as tf
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import (
    ArrayType,
    ByteType,
    DataType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
)

from kamae.spark.params import SingleInputSingleOutputParams
from kamae.spark.utils import single_input_single_output_array_transform
from kamae.tensorflow.layers import ArraySubtractMinimumLayer

from .base import BaseTransformer


class ArraySubtractMinimumParams(Params):
    """
    Mixin class containing pad value parameters needed
    for array subtract min transformers.
    """

    padValue = Param(
        Params._dummy(),
        "padValue",
        "The value to be considered as padding. Defaults to `None`.",
        typeConverter=TypeConverters.toFloat,
    )

    def setPadValue(self, value: float) -> "ArraySubtractMinimumParams":
        """
        Sets the parameter pad value to the given value.

        :param value: pad value.
        :returns: Instance of class mixed in.
        """
        return self._set(padValue=value)

    def getPadValue(self) -> float:
        """
        Gets the pad value parameter.

        :returns: float pad value.
        """
        return self.getOrDefault(self.padValue)


class ArraySubtractMinimumTransformer(
    BaseTransformer,
    SingleInputSingleOutputParams,
    ArraySubtractMinimumParams,
):
    """
    ArraySubtractMinimumTransformer that computes the difference within an array from
    the minimum non-paded element in the input tensor. The calculation preserves the pad
    value elements.

    The main use case in mind for this is working with an array of timestamps.
    """

    @keyword_only
    def __init__(
        self,
        inputCol: Optional[str] = None,
        outputCol: Optional[str] = None,
        inputDtype: Optional[str] = None,
        outputDtype: Optional[str] = None,
        layerName: Optional[str] = None,
        padValue: Optional[float] = None,
    ) -> None:
        """
        Initialise the ArraySubtractMinimumTransformer

        :param inputCol: Input column name.
        :param outputCol: Output column name.
        :param inputDtype: Input data type to cast input column to before
        transforming.
        :param outputDtype: Output data type to cast the output column to after
        transforming.
        :param layerName: Name of the layer. Used as the name of the tensorflow layer
        :param padValue: The value to be considered as padding. Defaults to `None`.
        :returns: None
        """
        super().__init__()
        self._setDefault(padValue=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @property
    def compatible_dtypes(self) -> Optional[List[DataType]]:
        """
        List of compatible data types for the layer.
        If the computation can be performed on any data type, return None.

        :returns: List of compatible data types for the layer.
        """
        return [
            FloatType(),
            DoubleType(),
            ByteType(),
            ShortType(),
            IntegerType(),
            LongType(),
        ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        """
        Performs the calculation of the differences on the input dataset.

        Example:
         dataset = spark.Dataframe(
            [
                [19, 18, 13, 11, 10, -1, -1, -1],
                [12, 2, 1, -1, -1, -1, -1, -1],
            ],
            'input_col'
         )
         Output: spark.Dataframe(
            [
                [19, 18, 13, 11, 10, -1, -1, -1],
                [12, 2, 1, -1, -1, -1, -1, -1],
            ],
            [
                [9, 8, 3, 1, 0, -1, -1, -1],
                [11, 1, 0, -1, -1, -1, -1, -1],
            ],
            'input_col', 'output_col'
        )

        :param dataset: The input dataframe.
        :returns: Transformed pyspark dataframe.
        """
        input_column_type = self.get_column_datatype(dataset, self.getInputCol())
        if not isinstance(input_column_type, ArrayType):
            raise ValueError(
                f"""Input column {self.getInputCol()} must be of ArrayType.
                Got {input_column_type} instead."""
            )
        padded_value = self.getPadValue()

        def array_subtract_min(x: Column, pad_value: Optional[float]) -> Column:
            if pad_value is None:
                return F.transform(x, lambda y: y - F.array_min(x))
            else:
                return F.transform(
                    x,
                    lambda y: F.when(
                        y != F.lit(pad_value),
                        y - F.array_min(F.filter(x, lambda z: z != F.lit(pad_value))),
                    ).otherwise(y),
                )

        array_subtract = single_input_single_output_array_transform(
            input_col=F.col(self.getInputCol()),
            input_col_datatype=input_column_type,
            func=lambda x: array_subtract_min(x, padded_value),
        )
        return dataset.withColumn(self.getOutputCol(), array_subtract)

    def get_tf_layer(self) -> tf.keras.layers.Layer:
        """
        Gets the tensorflow layer for the sequential difference transformer.

        :returns: Tensorflow keras layer with name equal to the layerName parameter that
        performs the sequential difference operation.
        """
        return ArraySubtractMinimumLayer(
            name=self.getLayerName(),
            input_dtype=self.getInputTFDtype(),
            output_dtype=self.getOutputTFDtype(),
            pad_value=self.getPadValue(),
        )



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

import tensorflow as tf

import kamae
from kamae.tensorflow.typing import Tensor
from kamae.tensorflow.utils import NormalizeLayer, enforce_single_tensor_input


@tf.keras.utils.register_keras_serializable(package=kamae.__name__)
class ConditionalStandardScaleLayer(NormalizeLayer):
    """
    Performs the standard scaling of the input with a masking condition.
    This layer will shift and scale inputs into a distribution centered around
    0 with standard deviation 1. It accomplishes this by precomputing the mean
    and variance of the data, and calling `(input - mean) / sqrt(var)` at
    runtime.
    The skip_zeros parameter allows to apply the standard scaling process
    only when input is not equal to zero. If equal to zero, it will remain zero in
    the output value as it was in the input value.
    """

    def __init__(
        self,
        mean,
        variance,
        name: str = None,
        axis=-1,
        input_dtype: str = None,
        output_dtype: str = None,
        skip_zeros: bool = False,
        epsilon: float = 0,
        **kwargs,
    ) -> None:
        """
        Intialise the ConditionalStandardScaleLayer layer.
        :param mean: The mean value(s) to use during normalization. The passed value(s)
        will be broadcast to the shape of the kept axes above; if the value(s)
        cannot be broadcast, an error will be raised when this layer's
        `build()` method is called.
        :param variance: The variance value(s) to use during normalization. The passed
        value(s) will be broadcast to the shape of the kept axes above; if the
        value(s) cannot be broadcast, an error will be raised when this
        layer's `build()` method is called.
        :param name: The name of the layer. Defaults to `None`.
        :param axis: Integer, tuple of integers, or None. The axis or axes that should
        have a separate mean and variance for each index in the shape. For
        example, if shape is `(None, 5)` and `axis=1`, the layer will track 5
        separate mean and variance values for the last axis. If `axis` is set
        to `None`, the layer will normalize all elements in the input by a
        scalar mean and variance. Defaults to -1, where the last axis of the
        input is assumed to be a feature dimension and is normalized per
        index. Note that in the specific case of batched scalar inputs where
        the only axis is the batch axis, the default will normalize each index
        in the batch separately. In this case, consider passing `axis=None`.
        :param skip_zeros: If True, in addition to the masking operation,
        do not apply the scaling when the values to scale are equal to zero.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        :param epsilon: Small value to add to conditional check of zeros. Valid only
        when skipZeros is True. Defaults to 1e-4.
        :param input_dtype: The dtype to cast the input to. Defaults to `None`.
        :param output_dtype: The dtype to cast the output to. Defaults to `None`.
        """
        super().__init__(
            name=name,
            input_dtype=input_dtype,
            output_dtype=output_dtype,
            mean=mean,
            variance=variance,
            axis=axis,
            **kwargs,
        )
        self.skip_zeros = skip_zeros
        self.epsilon = epsilon

    @enforce_single_tensor_input
    def _call(self, inputs: Tensor, **kwargs) -> Tensor:
        """
        Performs normalization on the input tensor(s) by calling the keras
        ConditionalStandardScaleLayer layer.
        It applies the scaling only to values matching the mask condition, if set.
        It applies the scaling only to values not equal to zero, if skip_zeros is set.
        Decorated with `@enforce_single_tensor_input` to ensure that
        the input is a single tensor. Raises an error if multiple tensors are passed
        in as an iterable.
        :param inputs: Input tensor to perform the normalization on.
        :returns: The input tensor with the normalization applied.
        """
        # Ensure mean and variance match input dtype.
        mean = self._cast(self.mean, inputs.dtype.name)
        variance = self._cast(self.variance, inputs.dtype.name)
        normalized_outputs = tf.math.divide_no_nan(
            tf.math.subtract(inputs, mean),
            tf.math.maximum(
                tf.sqrt(variance), tf.constant(self.epsilon, dtype=inputs.dtype)
            ),
        )
        # output is 0 if variance is 0
        normalized_outputs = tf.where(
            tf.equal(variance, 0),
            tf.zeros_like(normalized_outputs),
            normalized_outputs,
        )
        if self.skip_zeros:
            eps = tf.constant(self.epsilon, dtype=inputs.dtype)
            normalized_outputs = tf.where(
                tf.abs(inputs) <= eps,  # x = (0 +- eps)
                tf.zeros_like(normalized_outputs),
                normalized_outputs,
            )
        return normalized_outputs

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of the ConditionalStandardScaleLayer layer.
        Used for saving and loading from a model.
        Specifically adds additional parameters to the base configuration.
        :returns: Dictionary of the configuration of the layer.
        """
        config = super().get_config()
        config.update(
            {
                "skip_zeros": self.skip_zeros,
                "epsilon": self.epsilon,
            }
        )
        return config



# Copyright [2024] Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-ancestors
# pylint: disable=no-member
from typing import List, Tuple

import pyspark.sql.functions as F
from pyspark.ml.param import Params
from pyspark.sql import Column, DataFrame
from pyspark.sql.types import DataType


class InputOutputExtractor(Params):
    """
    Mixin class for extracting input & output information from a transformer/estimator.

    Used across all transformers/estimators to facilitate the construction
    of the pipeline graph.
    """

    def _get_single_or_multi_col(self, ingress: bool) -> List[str]:
        """
        Gets the input or output column name(s) of the layer. If the layer has a
        single input/output column, then the name of the column is returned as a
        singleton list. If the layer has multiple input/output columns, then the list of
        column names is returned.
        :param ingress: Whether to get input (True) or output (False) column names.
        :return: List of input or output column names.
        """
        single_prefix = "inputCol" if ingress else "outputCol"
        multi_prefix = f"{single_prefix}s"
        if self.hasParam(multi_prefix) and self.isDefined(multi_prefix):
            return self.getOrDefault(multi_prefix)
        elif self.hasParam(single_prefix) and self.isDefined(single_prefix):
            return [self.getOrDefault(single_prefix)]
        else:
            return []

    def get_layer_inputs_outputs(self) -> Tuple[List[str], List[str]]:
        """
        Gets the input & output information of the layer. Returns a tuple of lists,
        the first containing the input column names and the second containing the
        output column names.

        :returns: Tuple of lists containing the input and output column names.
        """
        inputs = self._get_single_or_multi_col(ingress=True)
        outputs = self._get_single_or_multi_col(ingress=False)
        return inputs, outputs

    @staticmethod
    def get_column_datatype(dataset: DataFrame, column_name: str) -> DataType:
        """
        Gets the datatype of a column in a dataset. First selects the column to ensure
        that if it is a struct type, the datatype of the struct element is returned.

        :param dataset: Dataset to get the column datatype from.
        :param column_name: Name of the column to get the datatype of.
        :return: Datatype of the column.
        """
        return (
            dataset.select(F.col(column_name).alias("input")).schema["input"].dataType
        )

    def get_multiple_input_cols(
        self, constant_param_name: str, input_cols_limit: int = None
    ) -> List[Column]:
        """
        Gets the (possibly multiple) input columns for the transformer.
        If using multiple input columns, we get them, raising an error if the number is
        greater than the limit. If using a single input column, we get the input column
        and the constant_param_name.

        :returns: List of columns
        """
        if self.isDefined("inputCols"):
            # If multiple input columns are defined, we get them, throwing an error
            # if there are more than the limit.
            input_cols = self.getOrDefault("inputCols")
            if input_cols_limit is not None and len(input_cols) > input_cols_limit:
                raise ValueError(
                    f"""
                    Number of input columns ({len(input_cols)})
                    exceeds limit ({input_cols_limit}).
                    """
                )

            return [F.col(c) for c in input_cols]
        elif (
            self.isDefined("inputCol")
            and self.getOrDefault(constant_param_name) is not None
        ):
            # If only one input column is defined, we use it
            # alongside the constant.
            return [
                F.col(self.getOrDefault("inputCol")),
                F.lit(self.getOrDefault(constant_param_name)).alias(
                    self.uid + constant_param_name
                ),
            ]
        else:
            # If neither inputCols nor inputCol & constant_param_name are defined,
            # we raise an error.
            raise ValueError(
                f"""
                Either inputCols or inputCol & {constant_param_name} must be defined.
                """
            )
